/***************************************************************************/
/*       Lotos, an extensible overlay network simulator                    */
/*       MNG Group                                                         */
/*       University of Virginia                                            */
/*       July, 2002                                                        */
/***************************************************************************/

package lotos;


import lotos.*; 

import lotos.event.SimEventMAC;
import lotos.event.SimEventMsg;
import lotos.event.SimEvent;

import java.util.*;
import java.lang.*;

import hypercast.*;
import hypercast.util.XmlUtil;

import org.apache.xpath.XPath;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
/**
 * This class simulate the 802.11 protocol under ad hoc mode
 * Currenly only the broadcast function (basic DCF access) is 
 * implemented
 * 
 * @author Guangyu Dong, Jan 6, 2004
 */
 
interface MAC_Callback{
	void messageSentFromMAC(SimEventMsg msg);
	void messageReceivedToMAC(SimEventMsg msg);
}

public class AdhocMAC implements I_MACInterface, I_Stats, I_EventHandler{
		
	/** Parameters used in 802.11 */
	/** Short Interframe Space */
	final static long SIFS=10;
	
	/** Distributed Interframe Space */
	final static long DIFS=50;
	/** The Backoff Time Slot */
	final static long SLOT_TIME=20;
	
	/** The Time needed to Sense the Signal */
	final static long SENSING_TIME=5;
	
	/** Maximum Contention Window Size */
	final static int CWmax=1023;
	/** Minimum Contention Window Size */
	final static int CWmin=31;
	
	/** Current Contention Window Size */
	int current_cw=31;
	
	/** Size of Outgoing Buffer*/
	final static int BUFFER_SIZE=1000;
	
	/** Outgoing Buffer */
	Vector msg_queue;
	
	/** the bandwidth in bytes/second */
	public int bandwidth;
	
	
	/** Packet Speed (m/s): distance between node/The time between a bit was sent out and it's received */
	final static float PACKET_SPEED=SimConfig.WIRELESS_PACKET_SPEED;
	/** Transmission Range in m */
	public float SIGNAL_RANGE = SimConfig.WIRELESS_DEFAULT_SIGNAL_RANGE;
	/** Signal Sensing Range in m */
	public float SENSE_RANGE = SimConfig.WIRELESS_DEFAULT_SENSE_RANGE;
	
	/** The Timer Index for Backoff Event */
	final static int BACKOFF_TIMER=1;
	/** The Timer Index for DIFS Event */
	final static int DIFS_TIMER=2;
	/** The Timer Index for Event Sending Out Packet */
	final static int SENDOUT_TIMER=3;
	
	/** statistics */
	long bytesSent;
	long bytesRec;
	long packetsSent;
	long packetsRec;
	
	
	
	/** The next DIFS Event.
	 *  Used to eliminate old events that has not been executed 
	 */
	SimEventMAC nextDIFS;
	/** The next Backoff Event 
	 *  Used to eliminate old events that has not been executed
	 */
	SimEventMAC nextBackoff;
	
	/** Callback object for sending out incoming message */
	MAC_Callback callback;
	
	/** the overlay network running the node*/
	PhysicalNetwork net;
	/** the node running this MAC protocol */
	PhysicalNode node;
	
	/** A random number generator */
	static Random ran;
	
	/* Constructor. 
	 * @param _node the PhysicalNode running this MAC
	 */
    public AdhocMAC(PhysicalNode _node, int bw){
		net=_node.net;
		node=_node;
		
		if(ran==null)
			ran=new Random();
		bandwidth = bw;
		
		MACInit();
		initStats();
	}
	/* 802.11 functions */
	
	/* @return current CW value */
	int getCW(){
		return current_cw;
	}
	
	boolean cw_set=false;
	/* Reset the current CW to the minimum value */
	void resetCW(){
		current_cw=CWmin;	
		
		//lotosDebug.errorln("reset CW="+getCW());
	}
	/* Calculate the next CW after detecting a contention */
	void nextCW(){
		int new_cw=(current_cw*2+1);
		if(new_cw<=CWmax)current_cw=new_cw;	
		
		//lotosDebug.errorln("next CW="+getCW());
	}
	
	int backoffTimer=0;
	/*
	 *  Backoff timer is initiated randomly and
	 *  decreased every slot time until it reaches 0
	 * 
	 *  @return return the next backoff timer 
	 */
	int getNextBackoffTimer(){
		if(backoffTimer<=0)backoffTimer=getRandomBackoffTime();
		else backoffTimer--;
		
		return backoffTimer;	
	}
	
	/*
	 * @return a randomly generated backoff time
	 */
	int getRandomBackoffTime(){
		//lotosDebug.errorln("CW="+getCW());
		return ran.nextInt(getCW()+1);	
	}
	/*
	 * Send out a message if the buffer is not empty
	 * 
	 * @return whether any message has been sent
	 */
	 
	boolean sendOutMessage(){
		
		
		if(msg_queue.size()>0){
			SimEventMsg msg=(SimEventMsg)msg_queue.elementAt(0);
			msg_queue.remove(0);
			
			//lotosDebug.errorln("Send out message");
			
			node.net.wl_phyLayer.sendMsg(msg);
			if(callback!=null)callback.messageSentFromMAC(msg);
			
			//sending success
			//set DIFS timers of all neighbors
			ms_timeLastSent = ms_timeLastReceived;
			resetCW();
			cw_set=false;
			packetsSent++;
			bytesSent+=msg.getSize();
			
			return true;
			
			
			
		}else {
			//lotosDebug.errorln("No message waiting");
		}
		return false;
	}
	/*
	 * @see MACInterface#getHeaderSize()
	 */
	public int getHeaderSize(){
		return 34;
	}
	
	long ms_timeLastReceived=0;
	long ms_timeLastSent=0;
	long ms_timeLastSensed=0;
	/* when a signal is sensed */
	public void senseMessage(long ms_time_start, long ms_time_end){
		if(ms_timeLastSensed<ms_time_end){
			ms_timeLastSensed = ms_time_end;
			insertDIFSEvent(ms_time_end-simTime());
		}
	}
	
	/*
	 * @see MACInterface#recordMessage(long, long)
	 */
	public void recordMessage(long ms_time_start, long ms_time_end){
		
		//lotosDebug.errorln("record message");
		
		if(ms_timeLastReceived > ms_time_start)
		{
			//lotosDebug.errorln("collision happend at receiver");
			//collision happens at the receiver
			//and the sender will not know
			//collision = true;
			
			if(ms_timeLastSent > ms_time_start){
				//sending fails
				//lotosDebug.errorln("Sending collision detected.");
			}
			
			
			node.wl_lastMsgCollision=true;
			if(ms_timeLastReceived<ms_time_end){
				ms_timeLastReceived = ms_time_end;
				insertDIFSEvent(ms_time_end-simTime());
			}
			//break;
		}else{
			//lotosDebug.errorln("no collisino");
			//record the message end time on receiver
			ms_timeLastReceived = ms_time_end;
			node.wl_lastMsgCollision = false;
			insertDIFSEvent(ms_time_end-simTime());
		}
		
	}
	
	/* handle the DIFS event */
	void handleDIFS(){
		
		long cf_time1=ms_timeLastReceived-simTime()+DIFS;
		long cf_time2=ms_timeLastSensed-simTime()+DIFS;
		
		if(cf_time1<=0 && cf_time2<=0){
			//no signal for a period of DIFS	
			//handle DIFS
			
			//lotosDebug.errorln("Medium is free");
			
			int bo_timer=getNextBackoffTimer();
			
			//lotosDebug.errorln("backoff timer "+bo_timer);
			
			if(bo_timer==0){
				//send message
				insertSendoutEvent();
				
			}else{
				//enter the backoff phase
				insertBackoffEvent();
			}
		}else{
			//lotosDebug.errorln("Medium is not free");
			
		}
		
	}
	/* handle the backoff event */
	void handleBackoff(){
		
		handleDIFS();
		
	}
	
	/* handle the SENDOUT event */
	void handleSendout(){
		sendOutMessage();
	}
	
	
	/* 
	 * insert the point that the sending ends and the backoff 
	 * should be started. 
	 * Both sender and receiver should be inserted this point.
	 */
	void insertDIFSEvent(long trans_time){
		//lotosDebug.errorln("Insert the next DIFS");
		nextDIFS=setMACTimer(trans_time+DIFS, DIFS_TIMER);
		
	}
	
	boolean idle=false;
	/*
	 * insert an backoff event happening at a slottime from now
	 */
	void insertBackoffEvent(){
		if(queueEmpty()){
			idle=true;
		}else{
			//lotosDebug.errorln("Insert the next backoff");
			nextBackoff=setMACTimer(SLOT_TIME, BACKOFF_TIMER);
			idle=false;
		}
	}
	/*
	 * Insert a event to send out message
	 */
	void insertSendoutEvent(){
		setMACTimer(SENSING_TIME, SENDOUT_TIMER);
	}
		
	/* message buffer functions */
	boolean queueEmpty(){
		return msg_queue.size()<=0;	
	}
	void insertMsg(SimEventMsg msg){
		if(msg_queue.size()<BUFFER_SIZE){
			msg_queue.addElement(msg);	
		}//else 
			//lotosDebug.errorln("message dropped due to full buffer");
	}
	SimEventMsg getNextPendingMsg(){
		if(msg_queue.size()>0){
			SimEventMsg msg=(SimEventMsg)msg_queue.elementAt(0);
			msg_queue.remove(0);
			return msg;
		}else return null;	
	}
	
	/*
	 * @return current time
	 */
	long simTime(){
		return net.simTime();
	}
	
	/*
	 * @see MACInterface#sendMsg(SimEventMsg)
	 */
	public boolean sendMsg(SimEventMsg msg){
		//debug
		////lotosDebug.errorln("insert a message");
		msg.setTime(simTime());
		
		insertMsg(msg);
		if(idle){
			idle=false;
			
			long channel_idle_time= (simTime()-(ms_timeLastReceived+DIFS));
			if(channel_idle_time>=0){
				long next_bo_time= SLOT_TIME - (channel_idle_time%SLOT_TIME) ;
				nextBackoff=setMACTimer(next_bo_time, BACKOFF_TIMER);
				//handleBackoff();
			}
		}
		
		return true;
	}
	
	/*
	 * @see MACInterface#MACInit()
	 */
	public void MACInit(){
		msg_queue=new Vector();
		
		nextDIFS = null;
		nextBackoff = null;
		current_cw=31;
		backoffTimer=0;
		idle=false;
		cw_set=false;
		
		ms_timeLastSent=-1;
		ms_timeLastReceived=-1;
		ms_timeLastSensed=-1;
		node.wl_lastMsgCollision=false;
		
		//check if some transmision is going around
		/*
		Vector neighbors = node.getWirelessNeighbors();
		if(neighbors!=null){
	   	   	for(int i=0;i<neighbors.size();i++) {
	   	   		PhysicalNode nnode = (PhysicalNode) neighbors.elementAt(i);
	   	   		AdhocMAC nmac= (AdhocMAC)nnode.wl_MAC;
				if(nmac.ms_timeLastSent>simTime()){
					if(ms_timeLastReceived>=0)node.wl_lastMsgCollision=true;
					if(ms_timeLastReceived < nmac.ms_timeLastSent)
						ms_timeLastReceived = nmac.ms_timeLastSent;
				}
			}
		}
		
		
	    //set DIFS timer
	    long trans_time;
	    if(ms_timeLastReceived>=0)trans_time=ms_timeLastReceived-simTime();
	    else trans_time=(long)(Math.random()*DIFS);
	    */
	    //insertDIFSEvent(trans_time + ran.nextInt(1000000));
	    insertDIFSEvent(ran.nextInt(1000000));
	    
	}
	/*
	 * @see MACInterface#MACStep()
	 */
	public void MACStep(){
		
		
	}
	/*
	 * @see MACInterface#handleMACTimer(SimEventMAC)
	 */
	public void handleEvent(SimEvent e){
		SimEventMAC mac_event = (SimEventMAC) e;
		int timer_index= mac_event.getIndex() ;
		
		switch(timer_index){
			case DIFS_TIMER:{
				if(mac_event==nextDIFS){
					//lotosDebug.errorln("DIFS event time="+simTime());
					handleDIFS();
				}
				break;
			}
			
			case BACKOFF_TIMER:{
				if(mac_event==nextBackoff){
					//lotosDebug.errorln("Backoff event time="+simTime());			
					handleBackoff();
					
				}
				break;
			}
			case SENDOUT_TIMER:{
				handleSendout();
				break;	
			}
			
			default:
				lotosDebug.errorln("Unknown MAC timer: "+timer_index);
		}
	}
	
	public long getSendTime(int msgsize) {
		return (long)(msgsize*SimEventProcessor.TIME_PER_SEC/bandwidth);
	}
	
	/** add by Guangyu
	 * set a MAC timer event
	 * @param interval the interval of this node
	 * @param index the type of the timer
	 *
	 */
	
	public SimEventMAC setMACTimer(long ms_interval, int index) {
		SimEventMAC timer = new SimEventMAC(SimEventProcessor.getSimTime() + ms_interval, this, index);
		SimEventProcessor.insertEvent(timer);
		return timer;
	}
	
	String statsName = "WirelessMAC";
	StatsProcessor statsPro;
	
	/**
     * statistics section
     * r values: packetsSent,packetsRec,bytesSent,bytesRec, usedBuffer
     * r/w values: bandwidth, bufferSize
     *
     */
    private void initStats() {
    	bytesSent = 0;
    	packetsSent = 0;
    	bytesRec = 0;
    	packetsRec = 0;
    	
    	statsPro = new StatsProcessor(this/* I_Stats object */, true/* readable flag */, true/* writable flag */);
    	statsPro.addStatsElement("TransmissionRange", new stat_transRange(), 1, 1);
    	statsPro.addStatsElement("SensingRange", new stat_senseRange(), 1, 1);
    	statsPro.addStatsElement("PacketsSent", new stat_packetsSent(), 1, 1);
    	statsPro.addStatsElement("BytesSent", new stat_bytesSent(), 1, 1);
    	statsPro.addStatsElement("PacketsRecieved", new stat_packetsRec(), 1, 1);
    	statsPro.addStatsElement("BytesRecieved", new stat_bytesRec(), 1, 1);
    	statsPro.addStatsElement("Bandwidth", new stat_bandwidth(), 1, 1);
    	/*
    	statsPro.addStatsElement("BufferSize", new stat_bufferSize(), 1, 1);
    	statsPro.addStatsElement("BufferUsed", new stat_usedBuffer(), 1, 1);
    	statsPro.addStatsElement("PacketsBuffered", new stat_packetsBuffed(), 1, 1);
    	*/
    }
    class stat_senseRange extends SimpleStats
	{
		
		protected String getStats()
		{
			return Float.toString(SENSE_RANGE);
		}

		protected String setStats(String newValue)
		{
			float newDist = Float.valueOf(newValue).floatValue();
			if(newDist>0) {
				SENSE_RANGE = newDist;
				net.updateWirelessAdjacency();
			}
			return Float.toString(SENSE_RANGE);
		}

		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Float"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
		public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException {
			return XmlUtil.createSchemaElement(doc, statisticsName, "xsd:Float", null, null);
		}
	}
    class stat_transRange extends SimpleStats
	{
		
		protected String getStats()
		{
			return Float.toString(SIGNAL_RANGE);
		}

		protected String setStats(String newValue)
		{
			float newDist = Float.valueOf(newValue).floatValue();
			if(newDist>0) {
				SIGNAL_RANGE = newDist;
				net.updateWirelessAdjacency();
			}
			return Float.toString(SIGNAL_RANGE);
		}

		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Float"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
		public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException {
			return XmlUtil.createSchemaElement(doc, statisticsName, "xsd:Float", null, null);
		}
	}
    class stat_bandwidth extends SimpleStats
	{
		
		protected String getStats()
		{
			return Integer.toString(bandwidth);
		}

		protected String setStats(String newValue)
		{
			int newBand = Integer.valueOf(newValue).intValue();
			if(newBand>0)
				bandwidth = newBand;
			
			return Integer.toString(bandwidth);	
		}

		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
		public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException {
			return XmlUtil.createSchemaElement(doc, statisticsName, "xsd:Integer", null, null);
		}
	}
    class stat_packetsRec extends SimpleStats{
		protected String getStats()
		{
			return Long.toString(packetsRec);
		}

		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}
    class stat_bytesRec extends SimpleStats{
		protected String getStats()
		{
			return Long.toString(bytesRec);
		}

		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}
    class stat_packetsSent extends SimpleStats{
		protected String getStats()
		{
			return Long.toString(packetsSent);
		}

		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}
    class stat_bytesSent extends SimpleStats{
		protected String getStats()
		{
			return Long.toString(bytesSent);
		}

		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}
    
    public String getStatsName() {
		return statsName;
	}
	public void setStatsName(String name) {
		statsName = name;
	}
	public Element[] getStats(Document doc, XPath xpath) throws HyperCastStatsException {
		return statsPro.getStatsResult(doc, xpath);
	}
	public Element[] setStats(Document doc, XPath xpath, Element newValue) throws HyperCastStatsException {
		return statsPro.setStatsResult(doc, xpath, newValue);
	}
	public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException {
		return statsPro.getReadSchemaResult(doc, xpath);
	}
	public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException {
		return statsPro.getWriteSchemaResult(doc, xpath);
	}
}
