package lotos;

import java.awt.Color;
import java.util.Enumeration;
import java.util.LinkedList;
import java.util.Vector;

import lotos.Protocol.NeighborsUpdateTimer;
import lotos.event.*;
import lotos.gateway.LotosGateway;
import lotos.history.ConsoleOutputHistoryEvent;

import hypercast.*;
import hypercast.events.NODE_ISSTABLE;
import hypercast.events.NODE_LEAVEOVERLAY;
import hypercast.events.NODE_LOGICALADDRESSCHANGED;
import hypercast.events.NODE_NEIGHBORHOODCHANGED;

import org.apache.xpath.XPath;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
/**
 * The LotosHyperCastAppl is an abstract application designed to run on a lotos physical node.
 * Users must implement the following in classes that descend from this one:
 * 1. Any constructor MUST start with a call to the default super constructor.
 * 2. This application can, and should, be ran as a seperate thread so that it doesn't block
 * the execution of the rest of the simulation. It is up to the user to implement the Runnable interface
 * and start the thread if the application uses threading. Simple fsm like applications can forgo 
 * using threading, and simply not implement Runnable.
 * 3. Time can accessed either using the provide method, currentTimeMillis, or directly by accessing the
 * SimEventProcessor class's static methods. Both methods access simulation time, not real time. Applications
 * accessing real time should use the System object.
 * 
 * 
 * Note: the LotosHyperCastAppl, like its predecessor, HyperCastAppl, cannot handle security.
 * 
 * @author Zhi An
 *
 */
public abstract class LotosHyperCastAppl extends HyperCastAppl implements I_ProtocolInterface {

	private LotosAdapter nodeAdapter;
	private LotosAdapter scktAdapter;
	private I_PhysicalInterface phy;
	private simHCNetwork network;
	
	public Color color = Color.green;
	
	private boolean joined=false;
	
	static HyperCastConfig TEMP_CONFIG;
	
	private LinkedList eventList = new LinkedList();
	private final static int MAXIMUM_SIZE_OF_EVENT_HISTORY = 300;
	
	private LinkedList consoleList = new LinkedList();
	private final static int MAXIMUM_SIZE_OF_CONSOLE = 300;
	
	/** neighbors in the overlay */
	private Vector logical_neighbors; // logical neighbors
	/** color used to display edges in the overlay */
	private Vector logical_colors; // the link color, display purpose
	/** the width of the logical edges */
	private Vector logical_width; // the link width, display purpose;
	/** neighbors which are external nodes */
	private Vector external_neighbors;
	
	Object neighborUpdateTimer;
	
	public LotosHyperCastAppl() {
		lotosDebug.errorln("Error creating application");
	}
	
	public LotosHyperCastAppl(HyperCastConfig config) {
		super(config);
		logical_neighbors = new Vector();
		logical_colors = new Vector();
		logical_width = new Vector();
		external_neighbors = new Vector();
		neighborUpdateTimer = new Object();

	}

	public void print(String str) {
		checkEventListSize();
		ConsoleOutputHistoryEvent entry = new ConsoleOutputHistoryEvent(SimEventProcessor.getSimTime(), getDetailedName(),str);
		getEventList().add(entry);
		if(SimConfig.log != null) {
			SimConfig.log.writeEvent(entry);
		}
		checkConsoleListSize();
		consoleList.add(new String(str));
		
	}
	
	public void println(String str) {
		print(str);
	}
	
	public String getName() {
		String label;
		I_LogicalAddress laddr = getLogicalAddress();
		label = "";
		if(network.displayNodeProtocol){
			label += ((PhysicalNode)phy).nodeInfo.application.name+"/"+((PhysicalNode)phy).nodeInfo.protocol.name;
			if(network.displayNodeLogicalAddress){
				if(laddr!=null && nodeAdapter.getPhysicalAddress()!=null)
					label += ":";
			}
		}
		if(network.displayNodeLogicalAddress){
			if(laddr!=null)
				label +=laddr.toString();
		}
		if(network.displayNodePhysicalAddress){
			//label +="@"+nodeAdapter.createPhysicalAddress();
			label+="@"+nodeAdapter.getPhysicalAddress();
		}
		if(network.displaySecuritySetting && ((PhysicalNode)phy).nodeInfo.security != null){
			label +=":" + ((PhysicalNode)phy).nodeInfo.security.name;
		}
		return label;
		
	}
	public String getDetailedName() {
		String label;
		I_LogicalAddress laddr = getLogicalAddress();
		label = "";
		
		label += ((PhysicalNode)phy).nodeInfo.application.name+"/"+((PhysicalNode)phy).nodeInfo.protocol.name;
			
		if(laddr!=null && nodeAdapter.getPhysicalAddress()!=null)
			label += ":";
	
		if(laddr!=null)
			label +=laddr.toString();
		
		label+="@"+nodeAdapter.getPhysicalAddress();
				
		return label;
	}

    //display properties
    /** the color of the logical node */
    public abstract Color getColor();
    
    /** the shape of the logical node */
    public abstract int getShape(); 

    /** handle message event 
     * when an application recieves a message, it simply passes that message to the appropriate
     * adapter, so that it arrives at the hypercast socket.
     * */
    public final void handleMsg( SimEventMsg msg ){
    	PhysicalNode src = msg.getSrc();
		if(src == null ) {
			lotosDebug.errorln("LotosHyperCastAppl: recieved message from unknown source");
		}
		
		//passes it to the adapter of the hypercast node
		//lotosDebug.println("sending msg to adapter");
		if(msg.msgType == SimEventMsg.PROTOCOL_MESSAGE)
			nodeAdapter.handleEvent(msg);
		else if (msg.msgType == SimEventMsg.APPLICATION_MESSAGE)
			scktAdapter.handleEvent(msg);
    }

    /**
     * this handles updating the logical links and returns events to the user by calling the receiveEvent function.
     */
    public final void handleEvent(SimEvent se) {
    	if(se instanceof SimEventTimer && ((SimEventTimer)se).getIndex() == neighborUpdateTimer && joined) {
    		if(SimConfig.calcLogical)
    			updateLogicalLinks();
    		se.setTime(SimEventProcessor.getSimTime() + SimConfig.neighborsUpdateTime);
			SimEventProcessor.insertEvent(se);
			//lotosDebug.println("updating logical links");
    	} else {
    		receiveEvent(se);
    	}
    }
    /**
     * a private helper function to update the logical links.
     *
     */
    private final void updateLogicalLinks () {
		//lotosDebug.println("Updating logical links at node "+adapter.getPhysicalAddress());
		//update logical links
		//clear all links
		clearLogicalEdges();
		//update links
		try {
			//lotosDebug.println("updating logical edges for "+adapter.getPhysicalAddress());
			I_LogicalAddress[] aps = getNeighbors();
			//lotosDebug.println("number of neighbors " + aps.length);
			//lotosDebug.println(aps.length + " neighbors");
			if (aps != null) {
				for (int i = 0; i < aps.length; i++) {				
					SimPhysicalAddress pa = (SimPhysicalAddress)network.getPhysicalAddress(aps[i]);
					
					PhysicalNode neighbor =	(PhysicalNode)LotosAdapterManager.searchAddress(pa);
					if (neighbor != null) {
						insertLogicalEdge(neighbor, Color.yellow, 2);
					} else {
						external_neighbors.add(pa);
						if(network.gateway_nodes.size() > 0)
							((LotosGateway)((PhysicalNode)network.gateway_nodes.get(0)).getProtocol()).fromLA = aps[i];
							
					}
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
    /**
     * this class should be implemented by the user for applications that use the sim event clock timers.
     * @param se the event
     */
    public abstract void receiveEvent(SimEvent se);
    
    /** start the protocol at this logical node */
    public final void startProtocol( ){
    	if(joined)
    		return;
    	joined = true;    		
    	//lotosDebug.println("lotos application joining overlay");
    	joinOverlay();
    	//start the timers for updating the logical links of this application
    	SimEventTimer timer = new SimEventTimer(SimEventProcessor.getSimTime() + SimConfig.neighborsUpdateTime, this, neighborUpdateTimer);
		SimEventProcessor.insertEvent(timer);
		
		start();
    }
    /**
     * this is a user defined method for starting the application.
     */
    public abstract void start();
    /** reset the protocol to its default state */
    public final void restartProtocol() {
    	if(!joined)
    		return;
    	try{
    		restart();
			clearLogicalEdges();
			leaveOverlay();
			this.nodeAdapter.Reset(null);
			this.scktAdapter.Reset(null);
			eventList.clear();
			joined = false;
		}catch(Exception e){
			e.printStackTrace();
		}
		
    }
    /** this is a user defined method for restarting the application. */
    public abstract void restart();
    /** cleanly stop the protocol */
    public final void stopProtocol() {
    	if(!joined) return;
    	    	
		try {
			stop();
			clearLogicalEdges();
			leaveOverlay();
			//lotosDebug.println("closing socket");
			closeSocket();
			nodeAdapter.Stop();
			scktAdapter.Stop();
			eventList.clear();
			
			joined = false;
		} catch (Exception e) {
			e.printStackTrace();
		}
		
    }
    /** this is a user-defined method for stopping the application */
    public abstract void stop();
    
    public void setNodeAdapter(LotosAdapter adpt) {
    	nodeAdapter = adpt;
    	phy = adpt.getPhyNode();
    }
    public void setScktAdapter(LotosAdapter adpt) {
    	scktAdapter = adpt;
    }
    
    public LotosAdapter getNodeAdapter() {
    	return nodeAdapter;
    }
    public LotosAdapter getScktAdapter() {
    	return scktAdapter;
    }
    public I_PhysicalAddress getPhysicalAddress() {
    	return nodeAdapter.createPhysicalAddress();
    }
    
    //TODO: change color with changing status.
    public void handle_NODE_NEIGHBORHOODCHANGED   (final NODE_NEIGHBORHOODCHANGED   event) { }
    public void handle_NODE_LEAVEOVERLAY          (final NODE_LEAVEOVERLAY          event) { }
    public void handle_NODE_ISSTABLE              (final NODE_ISSTABLE              event) { }
    
    public LinkedList getEventList() {
		return eventList;
	}
    public LinkedList getConsoleList() {
    	return consoleList;
    }
	public void checkEventListSize(){
		while(eventList.size()>=MAXIMUM_SIZE_OF_EVENT_HISTORY){
			eventList.removeFirst();
		}
	}
	public void checkConsoleListSize() {
		while(consoleList.size()>=MAXIMUM_SIZE_OF_EVENT_HISTORY){
			consoleList.removeFirst();
		}
	}
	
	/* default statistics */
	public String getStatsName ()
	{
		return "LotosApplication";
	}
	
	/**
	 * test if the node is its logical neighbor
	 */
	boolean hasLogicalEdge(PhysicalNode n) {
		if (logical_neighbors.contains(n))
			return true;
		else
			return false;
	}

	/**
	 * @return all the logical neighbors of this node
	 */
	public Enumeration getLogicalNeighbor() {
		return logical_neighbors.elements();
	}

	/**
	 * @return the colors of the logical links
	 */
	public Enumeration getLogicalColor() {
		return logical_colors.elements();
	}

	/**
	 * @return the lengths of the logical links
	 */
	public Enumeration getLogicalWidth() {
		return logical_width.elements();
	}

	/**
	 * @return the total number of logical neighbors
	 */
	int getLogicalNeighborNumber() {
		return logical_neighbors.size();
	}
	/**
	 * clear all the logical edges
	 */
	public void clearLogicalEdges() {
		this.logical_colors.clear();
		this.logical_neighbors.clear();
		this.logical_width.clear();
		this.external_neighbors.clear();
	}

	//physicalInterface
	/**
	 * part of the physicalInterface implemenation
	 * @return the logical neighbors of this node
	 */
	public Enumeration logicalLinks() {
		return getLogicalNeighbor();
	}

	/**
	 * insert a logical edge at this node
	 * @param n the other end of the logical edge
	 * @param color the color of the edge, for display purpose
	 * @param width the width of the edge, for dispaly purpose
	 */
	public boolean insertLogicalEdge(PhysicalNode n, Color color, int width) {
		if (logical_neighbors.contains(n))
			return false;
		else {
			logical_neighbors.addElement(n);
			logical_colors.addElement(color);
			logical_width.addElement(new Integer(width));
		}
		return true;
	}

	/**
	 * remove a logical neighbor
	 */
	public boolean removeLogicalEdge(PhysicalNode n) {
		if (logical_neighbors.contains(n)) {
			int index = logical_neighbors.indexOf(n);
			logical_neighbors.remove(index);
			logical_colors.remove(index);
			logical_width.remove(index);
			return true;
		} else {
			return false;
		}
	}
	public Vector getLogicalNeighbors() {
		return logical_neighbors;
	}
	public Vector getExternalNeighbors() {
		return external_neighbors;
	}

	public void setNetwork(simHCNetwork net) {
		network = net;
	}
	public long currentTimeMillis() {
		return SimEventProcessor.getSimTime()*1000;
	}
}
