/***************************************************************************/
/*       Lotos, an extensible overlay network simulator                    */
/*       MNG Group                                                         */
/*       University of Virginia                                            */
/*       July, 2002                                                        */
/***************************************************************************/
package lotos;

import lotos.*; 
import lotos.event.SimEvent;

import java.util.*;
import java.awt.*;
import java.awt.event.ActionListener;
import java.io.*;
import java.net.*;

import org.apache.xpath.XPath;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import hypercast.*;
import hypercast.util.XmlUtil;

/**
 * Overlay Network, the main class for the simulation.
 * Main components
 * a. physical nodes </p>
 *      the physical node </p>
 *      the protocol </p>
 *      the physical links and weights </p>
 *      the logical links </p>
 * b. a set of the active nodes within the network </p>
 * c. a discrete event simulator </p>
 * Main functions </p>
 * a. physical structure maintenence ( nodes add/deletion, physical link add/deletion) </p>
 * b. import/export physical topology </p>
 * c. import/export topology with protocol specific information </p>
 * d. execute the simulation, i.e. start, run, pause, step </p>
 * e. logical routing, used by the statisic collection package </p>
 * f. physical routing, used to provide services for the protocols </p>
 * g. graph-based simulation </p>
 *
 */
/**
 * revised by Guangyu Dong, Jan 6, 2004
 * to add mobility and wireless support
 */

public abstract class PhysicalNetwork implements I_Stats
{
	
	public static Vector LOCAL_ADDRESS;	
	
	long sum_sent=0;
	long sum_collision=0;
	long packet_sent=0;
	long packet_received=0;
	long bytes_sent=0;
	long bytes_received=0;
	
	public final static Color phyNodeColor = Color.black;
	public final static Color phyEdgeColor = Color.black;
	 final static float SIM_STEP = 0.005f;

	/** a vector that holds all the nodes in the network */
	public Vector nodes;
    /** a vector that holds all the wired edges in the network */
	public Vector edges;
    /** a vector that holds all the nodes in the overlay network */
	public Vector active_nodes;
	/** a vector that holds the nodes which connect to the internet. these nodes
	 * do not run hypercast.
	 */
	public Vector gateway_nodes;
    /** indicate whether a re-calculation of routing table is needed */
    boolean reroute=true;
    /** a discrete-event simulator */
    SimEventProcessor seq;
    /** The Physical Layer Simulator */
	//public PhysicalLayer phyLayer;
	public I_PhysicalLayer wl_phyLayer;
	///** a vector that holds all the physical edges in the network */
	//Vector edges;
	
	public int multicast_hop_limit = 255;
	
	public void setMulticastHopLimit(int limit){
		multicast_hop_limit = limit;
	}
	
	/** Nodes that are currently selected by mouse */
	protected Hashtable selectedNodes = new Hashtable();
	
	public int nextNodeID() {
		int low = 1;
		boolean found=true;
		for(low=1;low<=nodes.size()&&found;low++) {
			found=false;
			for(int i=0;i<nodes.size();i++) {
				PhysicalNode node = (PhysicalNode)nodes.get(i);
				if(node.id==low)
					found=true;
			}
			if(!found)
				low--;
		}
		return low;
	}
	
	public void setSelectedNodes(XYAddress start, XYAddress end, boolean append){
		if(!append){
			selectedNodes.clear();
		}
		for(int i=0;i<nodes.size();i++){
			PhysicalNode onode = (PhysicalNode)nodes.get(i);
			XYAddress addr = onode.getAddress();
			if(addr.x>=start.x && addr.y>=start.y && addr.x<=end.x && addr.y<=end.y){
				if(isNodeSelected(onode)){
					selectedNodes.remove(onode);
				}else{
					selectedNodes.put(onode, onode);
				}
			}
		}
	}
	
	public void setSelectedNode(PhysicalNode onode, boolean append) {
		if(!append){
			selectedNodes.clear();
		}
		if(onode!=null){
			if(isNodeSelected(onode)){
				selectedNodes.remove(onode);
			}else{
				selectedNodes.put(onode, onode);
			}
		}
	}
	public boolean isNodeSelected(PhysicalNode onode){
		return selectedNodes.containsKey(onode);
	}
	public PhysicalNode getFirstSelectedNode(){
		Enumeration e = selectedNodes.elements();
		if(e.hasMoreElements()){
			return (PhysicalNode)e.nextElement();
		}
		return null;
	}
	public PhysicalNode [] getSelectedNodes(){
		PhysicalNode [] allSelected = new PhysicalNode[selectedNodes.size()];
		Enumeration e = selectedNodes.elements();
		int i=0;
		while(e.hasMoreElements()){
			allSelected[i] = (PhysicalNode)e.nextElement();
			i++;
		}
		return allSelected;
	}
	
	public int getNumOfLeaf(){
		int num=0;
		if(active_nodes!=null){
			for(int i=0;i<active_nodes.size();i++){
				PhysicalNode onode=(PhysicalNode)active_nodes.get(i);
				if(onode.getProtocol().getLogicalNeighbors().size()<=1)num++;
			}
		}	
		return num;
	}

    /**
     * Constructor
     */
    PhysicalNetwork( ) {
		//setSimTime(0.0f);
		LotosAdapterManager.reset();
        //phyLayer = new SimplePhysicalLayer(this);
        wl_phyLayer = new WirelessPhysicalLayer(this);
        
        nodes = new Vector();
        edges = new Vector();
        active_nodes = new Vector();
        gateway_nodes = new Vector();
        seq = new SimEventProcessor();
        LOCAL_ADDRESS = new Vector();
        try {
        	Enumeration netinterfaces = NetworkInterface.getNetworkInterfaces();
        	while(netinterfaces.hasMoreElements()) {
        		Enumeration inetaddrs = ((NetworkInterface)netinterfaces.nextElement()).getInetAddresses();
        		while(inetaddrs.hasMoreElements())
        			LOCAL_ADDRESS.add(inetaddrs.nextElement());
        	}
    
        } catch(SocketException e) {
        	e.printStackTrace();    
        }
        initStats();
        //edges = new Vector();
    }
	
	/**
     * Build the physical layer routing table
     */
    public void buildPhysicalRoute(){
    	
        if( reroute == false ) return;
        reroute = false;
        // This function clears the unicast routing
        // tables and then calls rebuildNodeRoute
        // on each of the nodes to always build a
        // shortest path route.

        // clear away the old tables
        //lotusDebug.errorln("rebuilding routing table");
        
        try{
        	for (int i = 0; i < nodes.size(); i++)
            	((PhysicalNode) nodes.elementAt(i)).clearRoutingTable();

        	for (int i = 0; i < nodes.size(); i++)
        	{
	        	float diameter = rebuildNodeRoute((PhysicalNode) nodes.elementAt(i));
        	//if( i%200 == 0 ) lotusDebug.println("round:"+i);
	        /*if( i % 200 == 0 ) {
	            Runtime.getRuntime().gc();
	            lotusDebut.println("free memory:" + Runtime.getRuntime().freeMemory() );
	        }*/
	
		    }
		}catch(Exception e)
		{
			lotosDebug.errorln("Buidling physical route: "+e.toString());
		}

    }
	/**
     * build the routing table entries for a specific node
     */
    final float rebuildNodeRoute(PhysicalNode dst)
    {
        // This function rebuilds the route from every node
        // to the destination node.
        // Its a slower variant of Dijkstra's algorithm.
        // It assumes all edges are bidirectional.

        PriorityQueue_Dijkstra dedges = new PriorityQueue_Dijkstra(nodes.size());
        Hashtable hop_table = new Hashtable();


	//System.out.print("Routing To Node: ");
	//lotusDebut.println(dst.getAddress().toString());

        // insert nodes which are neighbors of the destination node.
        Enumeration neighbors = dst.getPhyNeighbor();
        int round=0;
        while (neighbors.hasMoreElements())
        {

            PhysicalNode n = (PhysicalNode) neighbors.nextElement();
            dedges.insert(dst.getDistanceTo(n), n);
            hop_table.put(n, dst);
        }

        float diameter = (float) 0.0;

        while (!dedges.isEmpty())
        {
            float distance_from_dst = dedges.minKey();
            PhysicalNode src = (PhysicalNode) dedges.removeMin();

            // if src does not already have a routing entry to the destination node.
            if (src.getPhyNextHop(dst) == null)
            {
                src.insertIntoRoutingTable(dst,(PhysicalNode)hop_table.get(src), distance_from_dst);
        		diameter = distance_from_dst;

                if (false)
                {
                    System.out.print("From ");
                    System.out.print(src.getAddress().toString());
                    System.out.print(" to ");
                    System.out.print(dst.getAddress().toString());
                    System.out.print(" via ");
                    System.out.println(((PhysicalNode)hop_table.get(src)).getAddress().toString());
                    System.out.print("; Sum distance =  ");
                    System.out.println(distance_from_dst);
                }

                // insert edges out of src
                if(!SimConfig.forwarding)continue; //don't need to calculate more than one hop
                
                neighbors = src.getPhyNeighbor();
                while (neighbors.hasMoreElements())
                {
                    PhysicalNode n = (PhysicalNode) neighbors.nextElement();

                    if (n == dst)
                        continue;

                    float distance = distance_from_dst + src.getDistanceTo(n);
                    if (dedges.getKeyOf(n) < 0)
                    {
                        dedges.insert(distance, n);
                        hop_table.put(n, src);
                    }
                    else if (distance < dedges.getKeyOf(n))
                    {
                        dedges.setKeyOf(n, distance);
                        hop_table.remove(n);
                        hop_table.put(n, src);
                    }
                }

            }
        }
        /*edges = null;
        hop_table = null;*/
        return diameter;
    }
    /**
     * add two symmetric physical edges between the nodes. edges are unidirectional.
     * the default bandwidth and buffer size are used.
     * @param from starting node
     * @param to   ending node
     * @param distance the lenght of the edge
     */
    public boolean addEdge( PhysicalNode from, PhysicalNode to, float distance ){
    	synchronized (edges) {
	    	//add edges in a symmetrically, i.e. always add two edges
	        if(from==to) return false;
	    	if( from.hasWiredEdge(to) ) return false;
			if( !from.wired_link || !to.wired_link)return false;
			
	        reroute = true;
	        WiredEdge edge = new WiredEdge(from, to, distance, SimConfig.WIRE_DEFAULT_BANDWIDTH, SimConfig.WIRE_DEFAULT_BUFFER);
	        WiredEdge rev = edge.reverse();
	        from.insertWiredEdge( to, edge );
	        to.insertWiredEdge( from, rev );
	        edges.add(edge);
	        edges.add(rev);
    	}
        return true;
    }
    /** 
     * adds one edge, used in the loading saved overlays
     * @param edge
     * @return
     */
    public boolean addEdge(WiredEdge edge) {
    	synchronized (edges) {
	    	PhysicalNode from = edge.srcNode;
	    	PhysicalNode to = edge.dstNode;
	    	if( from.hasWiredEdge(to) ) return false;
			if( !from.wired_link || !to.wired_link)return false;
			
	    	reroute = true;
	    	from.insertWiredEdge(to, edge);
	    	edges.add(edge);
    	}
    	return true;
    }
    /**
     * @return the vector containing the nodes in the overlay network
     */
    public Vector getActiveNodes() { return active_nodes;}

    /**
     * @return the vector containing all the nodes
     */
    public Vector getPhyNodes() { return nodes;}

    /*public boolean addLogicalEdge( PhysicalNode from, PhysicalNode to ){
	//add edges in a symmetrically way, i.e. always add two edges
        if( from.hasLogicalEdge(to) ) return false;

        from.insertLogicalEdge( to );
        to.insertLogicalEdge( from );
        return true;
    }*/

    /**
     * add a physical node with a specified address
     * @param addr the physical address of the node
     */
    public PhysicalNode addNode( XYAddress addr ){
        reroute = true;
	    Enumeration e = nodes.elements();
    	PhysicalNode onode;
	    boolean unique = true;

    	while( e.hasMoreElements() ){
            onode = (PhysicalNode)e.nextElement();
            if( onode.getAddress().equals(addr) ){
            lotosDebug.errorln(" non unique node insertion detected");
            lotosDebug.errorln(" non unique address:" + addr.getX() + ":" + addr.getY() );
            unique = false;
            break;
	        }
    	}

        onode = new PhysicalNode( addr, this,  null);
        onode.forwarding=SimConfig.forwarding;
        
        nodes.addElement( onode );
        
        if(onode.wireless_link)
        	updateWirelessAdjacency();
        statsPro.addStatsElement(onode.statsName, onode, 1, 1);
        return onode;

    }
	
	  /**
	   * added in Oct 25 for mobile node
     * add a physical node with a specified address
     * @param laddr the physical address of the node
     */
    public PhysicalNode addNode( PhyNodeProp pnode ){
    	    XYAddress addr = new XYAddress(pnode.GetX(),pnode.GetY());									   
	    Enumeration e = nodes.elements();
    	PhysicalNode onode;
	    boolean unique = true;
		

    	while( e.hasMoreElements() ){
            onode = (PhysicalNode)e.nextElement();
            if( onode.getAddress().equals(addr) ){
            lotosDebug.errorln(" non unique node insertion detected");
            lotosDebug.errorln(" non unique address:" + addr.getX() + ":" + addr.getY() );
            unique = false;
            break;
	        }
    	}

        onode = new PhysicalNode( pnode, this, null);
        onode.forwarding=SimConfig.forwarding;
        
        nodes.addElement( onode );
        statsPro.addStatsElement(onode.statsName, onode, 1, 1);
        return onode;

    }
    /**
     * remove a node in the network
     * @param node the node to be removed
     * @reutrn ture if the node exists, false otherwise
     */
    public boolean removeNode( PhysicalNode node ){
        if( nodes.contains(node) == false ) return false;
        reroute = true;
        Enumeration tmpEnum = node.getPhyNeighbor();
        while( tmpEnum.hasMoreElements() ) {
            PhysicalNode onode = (PhysicalNode) tmpEnum.nextElement();
            removeEdge(node, onode);
            
        }
        if(node.getProtocol()!=null) {
        	node.getProtocol().stopProtocol();
        }
        nodes.remove(node);
        active_nodes.remove(node);
        gateway_nodes.remove(node);
        updateWirelessAdjacency();
        
        hasMobileNode = false;
        for(int i=0;i<nodes.size();i++)
        	hasMobileNode = hasMobileNode || ((PhysicalNode)nodes.get(i)).isMobile();
        statsPro.removeStatsElement(node);
	    return true;
    }

    /**
     * remove an edge
     * @param from the starting node
     * @param to the ending node
     */
    public boolean removeEdge( PhysicalNode from, PhysicalNode to ){
    	synchronized(edges) {
	    	if( from.hasWiredEdge(to) == false || to.hasWiredEdge(from) == false ) return false;
	        from.removeWiredEdge(to);
	        to.removeWiredEdge(from);
	        reroute = true;
	        
	        for(int i = 0;i<edges.size();i++) {
	        	WiredEdge edge = (WiredEdge)edges.get(i);
	        	if((edge.srcNode == from && edge.dstNode == to) || (edge.dstNode == from && edge.srcNode==to)) {
	        		if(edges.remove(edge)) {
	        			i--;
	        			//lotosDebug.println("removing edge btwn " + nodes.indexOf(edge.srcNode) + " and "+ nodes.indexOf(edge.dstNode));
	        		}
	        	}
	        }
    	}
	    return true;
    }

    /**
     * An abstract function, which must be implemented by each overlay
     * protocol.
     * @param src the source node
     * @param dst the destination node
     * @return the next hope node in the logical network
     */
    public abstract PhysicalNode getLogicalNextHop(PhysicalNode src, PhysicalNode dst);
	
	
	
    /**
     * find the closest node to a physical location. Used in GUI interface
     */
    public PhysicalNode findClosest( XYAddress addr, int radius ){
        int i;
        float dist=Float.MAX_VALUE;
        int index=-1;

        for( i=0; i<nodes.size(); i++){
            if( dist > addr.distanceTo( ((PhysicalNode)nodes.elementAt(i)).getAddress()  ) ){
            dist = addr.distanceTo( ((PhysicalNode)nodes.elementAt(i)).getAddress()  );
            index=i;
            }
        }

        if( -1 == index ) return null;
        if( dist > radius ) return null;

        return (PhysicalNode)nodes.elementAt(index);

    }

    /**
     * find the edge which is the closest to a physical address
     */

    public boolean findClosestLine( XYAddress addr, int radius, PhysicalNode []nodeArray ){
        double dista;
        double distb;
        double distc;
        double cos1;
        double cos2;
        double sin1;

        for( int i=0; i<nodes.size(); i++){
            for( int j=0; j<i; j++){
                PhysicalNode fromnode = (PhysicalNode) nodes.elementAt(i);
                PhysicalNode tonode = (PhysicalNode) nodes.elementAt(j);

                dista = addr.distanceTo(fromnode.getAddress());
                distb = addr.distanceTo(tonode.getAddress());
                distc = fromnode.getAddress().distanceTo(tonode.getAddress());

                /* the structure

                        from ------------c-------------to
                           - 1                       2-
                             -                      -
                               a                 b
                                 -            -
                                    -      -
                                       node

                    */
                cos1 = (dista*dista + distc*distc - distb*distb)/(2*dista*distc);
                cos2 = (distb*distb + distc*distc - dista*dista)/(2*distb*distc);

                if( cos1 < 0 || cos2 < 0 ) continue;
                sin1 = Math.sqrt(1.0 - cos1*cos1);
                if( dista*sin1 <= radius ) {
                    if( fromnode.hasWiredEdge(tonode) ) {
                        nodeArray[0] = fromnode;
                        nodeArray[1] = tonode;
                        return true;
                    }
                }
            }
        }

        return false;
    }

    /**
     * @return the physical topology of the network
     *//*
    public TopoAdaptor returnTopo( ){
	    TopoAdaptor_HM topo = new TopoAdaptor_HM();

	    topo.nodes = new PhyNodeProp[nodes.size()];

        int numedges=0;
        Enumeration e = nodes.elements();

        while( e.hasMoreElements() ){
            PhysicalNode onode =(PhysicalNode) e.nextElement();
            numedges += onode.getWiredNeighbors().size();
        }

        topo.edges = new WiredEdge[numedges];

        int i, j;
        for( i=0,j=0; i<nodes.size(); i++){

            PhysicalNode fromnode = (PhysicalNode) nodes.elementAt(i);
            
            topo.nodes[i] = new PhyNodeProp(
                        fromnode.getAddress().getX(),
                        fromnode.getAddress().getY(),
                        fromnode.getMobilityStr(),
                        fromnode.mobile_speed,
                        fromnode.isWired(),
                        fromnode.isWireless() );

            /*if( active_nodes.contains( fromnode )){
            topo.activeNodes.addElement( topo.nodes[i] );
            }

            Enumeration neighbor = fromnode.getWiredNeighbors().elements();

            while( neighbor.hasMoreElements() ){
            PhysicalNode tonode = (PhysicalNode) neighbor.nextElement();

            topo.edges[j++] = new WiredEdge(
                            i,
                            nodes.indexOf(tonode),
                            fromnode.getDistanceTo(tonode), 1000000, 50000);

            }
        }

        return topo;
    }
*/
	/**
	 * Update the Adjacency list of node to node connectivity
	 */
	public void updateWirelessAdjacency()
	{
		reroute=true;
		for (int i = 0; i < nodes.size(); i++)
			((PhysicalNode) nodes.elementAt(i)).updateWirelessNeighbors();
	}
	public void moveMobileNodes()
	{
		reroute=true;
		for (int i = 0; i < nodes.size(); i++)
			((PhysicalNode) nodes.elementAt(i)).move();
	}
	/* deprecated 08/05/2006 - zhi
	public void MACStep()
	{
		reroute=true;
		for (int i = 0; i < nodes.size(); i++)
			((PhysicalNode) nodes.elementAt(i)).MAC.MACStep();
	}
	*/
    /**
     * @return the simulation time
     */
	public final long simTime(){ return seq.getSimTime(); }

    /**
     * insert an simulation event to the discrete event simulator
     */
    final void insertEvent( SimEvent e ) { seq.insertEvent(e); }
	final void clearEvent(SimEvent e) {seq.clearEvent(e);};
	final long getEvent(SimEvent e) {return seq.getEvent(e);}
	
	
    /**
     * get all the pending events
     */
	public final Enumeration getPendingEvent( ) { return seq.getPendingEvents(); }

   /**
    * this function checks all the physical links for messages that have been finished
    * being transmitted. such messages are removed and placed into the sim event processor.
    */
   protected void checkPhyLinks() {
	   int i;
	   synchronized (edges) {
		   for(i=0;i<edges.size();i++) {
		   
			   WiredEdge edge = (WiredEdge)edges.get(i);
			   //check if the edge isn't idle and it has a message that has finished being transmitted
			   while(edge.finishSending > 0 && edge.finishSending < seq.getSimTime()) {
				   SimEvent msg = edge.getMessage();
				   seq.insertEvent(msg);
			   }				   
		   }
	   }
   }
    //start the simulation: start the timers of each node and sending the Iam leader to
    //the server
    boolean simRunning = false;

    /**
     * Abstract function, used by each overlay protocol to implement those
     * initialze functions before the staring of the simulation
     */
    abstract boolean start();
    /**
     * Abstract function, used by each overlay protocol to clean up the current simulation
     * and restore the simulator to its default state
     */
    void restart(){
    	seq.reset();
		
    	for(int i=0;i<this.nodes.size();i++){
    		PhysicalNode onode = (PhysicalNode)nodes.elementAt(i);
    		//if(!gateway_nodes.contains(onode))
    		//onode.MAC.MACInit();
    			if(onode.wl_MAC != null)
    				onode.wl_MAC.MACInit();
    		onode.resetWiredEdge();
    	}
    	
    }
    /**
     * @return whether the simulation is running
     */
    public boolean isRunning() { return simRunning; }

    /**
     * An abstract function, implemented by each overlay protocol to build
     * its own overaly topology. This function actually implements the
     * static simulation
     */
    //abstract boolean buildOverlay();

    //the protocol-dependent part of the GUI interface part
    /**
     * Abstract function, used by each protocol to implement protoc-specific
     * functions
     */
    public abstract PopupMenu getMenus( );
    
   
    public static PopupMenu convertStringsToMenu(String[] menuName, ActionListener al){
    	PopupMenu pmenu = new PopupMenu();
    	MenuItem []mi_sim = new MenuItem[menuName.length];
        for( int i=0; i<menuName.length; i++){
        	mi_sim[i] = new MenuItem( menuName[i] );
            pmenu.add(mi_sim[i]);
            mi_sim[i].addActionListener(al);
        }
        return pmenu;
    }
    public static int getIndexOfMenu(String menu, String[] menu_names){
    	for(int i=0;i<menu_names.length;i++){
    		if(menu_names[i].equals(menu))
    			return i;
    	}
    	return -1;
    }
    abstract public boolean activateNode( PhysicalNode o );
    abstract public boolean deactivateNode(PhysicalNode o);
    
    public void setSpeed(float speed){
    	for(int i=0;i<nodes.size();i++){
    		PhysicalNode onode = (PhysicalNode)nodes.get(i);
    		onode.setMobileSpeed(speed);
    	}
    }
    /**
     * Abstract function, it is the callback function invoked by the simulator
     * and implemented by each overlay protocol to execute protocol-specific
     * operations
     */
    //public abstract boolean fireMenu( PhysicalNode onode, int index );

    
	//the protocol menu and instance creation part
    final static String [] theMobilemode = {
          "Random Motion",
		  "Linear Motion",
    };
    
    public boolean hasMobileNode = false;
    /** @return current supported protocols */
    public static String[] getMobilemode(){
	    return theMobilemode;
    }

    //import/export protocol-specific protocols
    /**
     * An abstract function, implemented by each protocol to read topology description file
     * with protocol-specific information
     */
    abstract boolean protocolReader( String fname );
    
    /**
     * An abstract function, implemented by each protocol to save topology description file
     * with protocol-specific information
     */
    abstract boolean protocolWriter( String fname );
    
    String statsName = "PhysicalNetwork";
    StatsProcessor statsPro;
    private void initStats() {
    	statsPro = new StatsProcessor(this/* I_Stats object */, true/* readable flag */, true/* writable flag */);
    	statsPro.addStatsElement("PacketSpeed", new stat_packetSpeed(), 1, 1);
    }
    class stat_packetSpeed extends SimpleStats{
		protected String getStats()
		{
			return Float.toString(SimConfig.WIRE_PACKET_SPEED);
		}

		protected String setStats(String newValue)
		{
			/*
			 * this only affects new packets. therefore, it should only be 
			 * set before the simulation starts.
			 */
			float speed = Float.parseFloat(newValue);
			if(speed>0)
				SimConfig.WIRE_PACKET_SPEED = Float.parseFloat(newValue);
			
			return newValue;
			
		}

		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Float"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
		public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException {
			return XmlUtil.createSchemaElement(doc, statisticsName, "xsd:Float", null, null);
		}
	}
    public String getStatsName() {
		return statsName;
	}
	public void setStatsName(String name) {
		statsName = name;
	}
	public Element[] getStats(Document doc, XPath xpath) throws HyperCastStatsException {
		return statsPro.getStatsResult(doc, xpath);
	}
	public Element[] setStats(Document doc, XPath xpath, Element newValue) throws HyperCastStatsException {
		return statsPro.setStatsResult(doc, xpath, newValue);
	}
	public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException {
		return statsPro.getReadSchemaResult(doc, xpath);
	}
	public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException {
		return statsPro.getWriteSchemaResult(doc, xpath);
	}
}

