/***************************************************************************/
/*       Lotos, an extensible overlay network simulator                    */
/*       MNG Group                                                         */
/*       University of Virginia                                            */
/*       July, 2002                                                        */
/***************************************************************************/
package lotos;

import lotos.*; 
import lotos.WiredEdge;
import lotos.event.*;

import java.awt.Color;
import java.util.Enumeration;
import java.util.Vector;

import hypercast.HyperCastStatsException;
import hypercast.I_Stats;
import hypercast.StatsProcessor;
import hypercast.util.XmlUtil;
import hypercast.SimpleStats;
import hypercast.I_Node;
//import hypercast.StatsElement;

import org.apache.xpath.XPath;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * Overlay Node, composed of physical and logical parts
 * Main components </p>
 * a. physical address </p>
 * b. physical links </p>
 * c. logical links </p>
 * d. physical name </p>
 * e. routing table to provide the physical network service </p>
 * f. protocol(logical part) </p>
 * Main functions </p>
 * a. maintenence of the physical/logical links </p>
 *      note: the physical links are added/deleted by outside
 *            the logical links are added/deleted by the protocol/graph-sim
 * b. dispatch/receive the events for two layers. </p>
 *          interface to the event dispatcher: send/receive the event(timer/msg)
 *          interface to the protocol: send/receive the event(timer/msg)
 *          for the message, it may need to route the message to the next hop
 * c. add/delete/start/end the running protocol </p>
 */
/**
 * revised by Guangyu Dong, Jan 6, 2004
 * to add mobility and wireless support
 */

public class PhysicalNode implements I_PhysicalInterface, I_Stats, I_EventHandler {
	/** id, used for determining the "IP address" of this physical node within lotos" */
	int id;
	/** Statistics - combining of underlay and overlay statistics */
	String statsName = "PhysicalNode";
	StatsProcessor statsPro;
	/** if a hc node has been inserted into statistics */
	//boolean nodestats=false;
	
	NodeInfo nodeInfo;
	
	/** Start of addon by Guangyu to support mobility and wireless network*/
	/** Mobility Mode */
	I_MobilityInterface mobility;

	/** There's a wired link? */
	public boolean wired_link = true;
	/** There's a wireless link? */
	public boolean wireless_link = false;

	/** Default moving speed */
	public static final float DEF_MOBILE_SPEED = 0.2f;

	/** wireless transmission range */
	//float wireless_trans_radius = WirelessPhysicalLayer.SIGNAL_RANGE;
	/** wireless sensing range */
	//float wireless_sense_radius = WirelessPhysicalLayer.SENSE_RANGE;
	/** moving speed of this node */
	float mobile_speed = DEF_MOBILE_SPEED;

	/** neighbors inside the wireless transmission range */
	Vector wireless_neighbors;
	/** distances to wireless neighbors */
	Vector wireless_distances;

	/** neighbors inside the wireless sensing range */
	Vector wireless_sensing_neighbors;


	/* Forwarding message??
	 * This flag decided whether this network is a multihop network
	 * If yes, every node will work as a rounter.
	 */
	boolean forwarding = true;

	/**
	 * @return whether the node has wireless link
	 */
	public boolean isWireless() {
		return wireless_link;
	}
	/**
	 * @return whether the node has wired link
	 */
	public boolean isWired() {
		return wired_link;
	}
	/**
	 * @return ths string representation of the mobility mode
	 */
	public String getMobilityStr() {
		if (mobility == null)
			return null;
		else
			return mobility.toString();
	}
	/*
	 * Set the mobility mode
	 */
	public void setMobility(String mobilityMode) {
		//lotusDebug.errorln("set mobility "+mobilityMode);
		setMobility(mobilityMode, DEF_MOBILE_SPEED);
		//lotusDebug.errorln("mobility= "+mobility);
	}
	/**
	 * Set the mobility mode and the speed
	 */
	public void setMobility(String mobilityMode, float speed) {

		if (mobilityMode == null) {
			mobility = null;
			mobile_speed = 0;
			return;
		}
		//lotosDebug.println("set mobility "+mobilityMode);
		//lotusDebug.errorln("mobile_speed="+mobile_speed);
		//lotusDebug.errorln("speed="+speed);
		mobile_speed = speed;

		if (mobilityMode.equals("Random Motion"))
			mobility = new MobiRandomStep(speed);
		else if (mobilityMode.equals("Linear Motion"))
			mobility = new MobiReflectiveLinear(speed);
		else {
			mobility = null;
		}
	}

	/*
	 * @return if this node is mobile
	 */
	public boolean isMobile() {
		return mobility != null;
	}
	/*
	 * set the speed
	 */
	public void setMobileSpeed(float sp) {
		mobile_speed = sp;
		if(this.mobility!=null)
			setMobility(this.mobility.toString(), sp);
	}
	/*
	 * set wireless transmission range
	 */
	/*
	public void setRadius(float tr) {
		wireless_trans_radius = tr;
	}
	*/
	/*
	 * @return wireless transmission range
	 */
	public float GetRadiusF() {
		return wl_MAC.SIGNAL_RANGE;
	}
	/*
	 * @return integer value of wireless transmission range
	 */
	public int GetRadius() {
		return (int)wl_MAC.SIGNAL_RANGE;
	}
	/*
	 * Set link types
	 * @param wd wired link flag
	 *        wl wireless link flag
	 */
	private void setLinkType(boolean wd, boolean wl) {
		wired_link = wd;
		wireless_link = wl;

		if (!wired_link) {
			//clear neighbors
			physical_neighbors = new Vector();
			physical_distances = new Vector();
			physical_edges = new Vector();

		}
		if (!wireless_link) {
			//clear neighbors
			wireless_neighbors = new Vector();
			wireless_distances = new Vector();

			wireless_sensing_neighbors = new Vector();
		}
	}
	/*
	 * Set link type
	 * @param type 
	 */
	public void setLinkType(int type) {
		switch (type) {
			case INTERFACE_WIRED :
				setLinkType(true, false);
				break;
			case INTERFACE_WIRELESS :
				setLinkType(false, true);
				break;
			case INTERFACE_BOTH :
				setLinkType(true, true);
				break;
		}
	}

	/*
	 * Move the node for 1 second
	 */
	public void move() {
		if (!isMobile())
			return;

		if (mobility != null) {
			Position prev = new Position(addr.getXF(), addr.getYF(), 0);
			Position next = mobility.move(net.simTime()/SimEventProcessor.TIME_PER_SEC, prev);
			moveTo(next.getX(), next.getY());
		}
	}
	/*
	 * Move to a specific location
	 */
	public void moveTo(float x, float y) {
		//lotusDebug.errorln(""+this+"move to "+x+","+y);
		addr = new XYAddress(x, y);
	}
	/**
	 * @return geographic distance to another node
	 */
	public final float getDirectDistanceTo(PhysicalNode n) {
		if (n == this)
			return (float)0.0;

		return (float)Math.sqrt(
			Math.pow(n.addr.x - addr.x, 2) + Math.pow(n.addr.y - addr.y, 2));
	}
	
	public final float getWiredDistanceTo(PhysicalNode n) {
		int index = physical_neighbors.indexOf(n);
		if(index<0)		
			return getDirectDistanceTo(n);
		return ((WiredEdge)physical_edges.get(index)).distance;
	}
	
	/**
	 * get the multi-hop netowrk distance to a node
	 * @param n the destination node
	 * @return the distance
	 */
	public final float getDistanceTo(PhysicalNode n) {
		if (n == this)
			return (float)0.0;

		int index = physical_neighbors.indexOf(n);
		if (index != -1)
			return ((Float)physical_distances.elementAt(index)).floatValue();
		else if (isWirelessNeighbor(n))
			return getDirectDistanceTo(n);

		if (routing_nexthop == null)
			return -1.0f;
		index = net.getPhyNodes().indexOf(n);
		if (index != -1) {
			float dist = 0;
			PhysicalNode nextnode = routing_nexthop[index];
			if (null == nextnode)
				return -1.0f;
			PhysicalNode prevnode = this;
			while (nextnode != n) {
				dist += prevnode.getDistanceTo(nextnode);
				prevnode = nextnode;
				nextnode = nextnode.getPhyNextHop(n);
			}
			dist += prevnode.getDirectDistanceTo(nextnode);
			return dist;
		} else {
			lotosDebug.errorln("can not route between the two points");
			return -1.0f;
		}

		//return Float.MAX_VALUE;

	}

	/*
	 * @return if otherNode is a wired network neighbor 
	 */
	boolean isWiredNeighbor(PhysicalNode otherNode) {
		if (!wired_link)
			return false;
		if (!otherNode.wired_link)
			return false;

		return physical_neighbors.indexOf(otherNode) >= 0;
	}
	/*
	 * @return if otherNode is inside the sensing range
	 */
	boolean isSensingNeighbor(PhysicalNode otherNode) {
		if (!wireless_link)
			return false;
		if (!otherNode.wireless_link)
			return false;

		float distance = getDirectDistanceTo(otherNode);
		if ((distance <= wl_MAC.SENSE_RANGE))
			return true;
		else
			return false;
	}
	/*
	 * @return if otherNode is inside the wireless transmission range
	 */
	boolean isWirelessNeighbor(PhysicalNode otherNode) {
		if (!wireless_link)
			return false;
		if (!otherNode.wireless_link)
			return false;

		float distance = getDirectDistanceTo(otherNode);
		if ((distance <= wl_MAC.SIGNAL_RANGE))
			return true;
		else
			return false;
	}
	/*
	 * @return if otherNode is a network neighbor
	 */
	boolean isNeighbor(PhysicalNode otherNode) {

		if (otherNode == this)
			return true;

		return isWiredNeighbor(otherNode) || isWirelessNeighbor(otherNode);

	}

	/**
	 * insert a wireless edge for this node
	 * @param n the new neigbhor
	 * @param weight the length of the node
	 */
	boolean insertWirelessEdge(PhysicalNode n, float weight) {
		if (!wireless_link)
			return false;

		if (wireless_neighbors.contains(n))
			return false;
		else {
			wireless_neighbors.addElement(n);
			wireless_distances.addElement(new Float(weight));
		}
		return true;
	}
	/**
	 * remove an wireless edge
	 * @param n the node at the other end of the edge
	 * @return true if such a neighbor exists, false otherwise
	 */
	boolean removeWirelessEdge(PhysicalNode n) {
		if (wireless_neighbors.contains(n)) {
			int index = wireless_neighbors.indexOf(n);
			wireless_neighbors.remove(index);
			wireless_distances.remove(index);
			return true;
		} else {
			return false;
		}
	}

	/*
	 * @return all wired neighbors
	 */
	public Vector getWiredNeighbors() {
		return physical_neighbors;
	}
	/*
	 * @return physical edge length to otherNode
	 */
	public float getWiredEdgeLen(PhysicalNode otherNode) {
		int index = getAllNeighbors().indexOf(otherNode);
		if (index >= 0)
			return getWiredEdgeLen(index);
		else
			return -1;
	}
	/*
	 * @return wired edge length of some neighbor
	 */
	public float getWiredEdgeLen(int index) {
		return ((Float)physical_distances.elementAt(index)).floatValue();
	}
	/*
	 * @return all wireless neighbors
	 */
	public Vector getWirelessNeighbors() {
		return wireless_neighbors;
	}
	/*
	 * @return all nodes inside the wireless sensing range 
	 */
	public Vector getSensingNeighbors() {
		return wireless_sensing_neighbors;
	}

	/** @return all neighbors, including wired and wireless neighbors */
	public Vector getAllNeighbors() {
		Vector all = new Vector();
		if (wired_link)
			all.addAll(physical_neighbors);
		if (SimConfig.AD_HOC_ROUTING && wireless_link)
			all.addAll(wireless_neighbors);

		return all;
	}
	/** @return all distances */
	public Vector getAllDistances() {
		Vector all = new Vector();
		if (wired_link)
			all.addAll(physical_distances);
		if (wireless_link)
			all.addAll(wireless_distances);

		return all;
	}
	/**
	 * Update the wireless neighbor list
	 */
	public void updateWirelessNeighbors() {

		if (!wireless_link)
			return;

		/*
		float distance;
		
		Enumeration enum = wireless_neighbors.elements();
		while( enum.hasMoreElements() ) {
		   otherNode = (PhysicalNode) enum.nextElement();
		   removeWirelessEdge(otherNode);
		}
		*/
		PhysicalNode otherNode;

		//clear the wireless neighbor
		wireless_neighbors.clear();
		wireless_distances.clear();
		wireless_sensing_neighbors.clear();

		for (int i = 0; i < net.nodes.size(); i++) {
			otherNode = (PhysicalNode)net.nodes.elementAt(i);
			if (isWirelessNeighbor(otherNode))
				insertWirelessEdge(otherNode, getDirectDistanceTo(otherNode));
			else if (isSensingNeighbor(otherNode))
				wireless_sensing_neighbors.addElement(otherNode);
		}
	}

	/** predefined link type */
	public static final int INTERFACE_WIRED = 0;
	public static final int INTERFACE_WIRELESS = 1;
	public static final int INTERFACE_BOTH = 2;

	/** End of addon by Guangyu */

	/** node attributes */
	/** name of the node */
	String name;

	/** physical node*/
	//disable by Guangyu
	//because it's not good to use it here
	//PhyNodeProp phyNode;

	/** physical address of the node
	 *  it's actually the location on the GUI 
	 */
	XYAddress addr;

	//logical structure
	/**
	 * protocol running on this node. It can be empty
	 * if no protocol is running
	 */
	I_ProtocolInterface protocol;

	//physical structure
	/** physical neighbors */
	Vector physical_neighbors; // physical neighbors
	/** the lenght of the physical edges */
	Vector physical_distances; // physical neighbors
	/** the physical edges */
	public Vector physical_edges;

	//routing table, the next hop is indexed by the node's position
	//in the PhysicalNetwork's nodes vector
	/** routing table at this node. It stores the next hope node from this
	 *  node to all the other nodes
	 *
	 */
	PhysicalNode[] routing_nexthop;
	/** a referece to the class Overlay Network */
	PhysicalNetwork net;

	/** MAC Layer simulator */
	
	//public MACInterface MAC; it appears that this is never used.
	public AdhocMAC wl_MAC;

	/** time when the last message sent by this node will be delivered (junk data) */
	public float timeLastMsg;

	/** last time this node stop sending or receving data */
	//public float wl_timeLastSent=-10f;
	//public float wl_timeLastReceived=-10f;
	public boolean wl_lastMsgCollision = false;

	/**constructor
	 * used when loading a saved network
	 * @param a node's physical address
	 * @param onet a reference to the overlay network
	 * @param sname node's name
	 */
	public PhysicalNode(XYAddress a, PhysicalNetwork onet, String sname) {
		//phyNode = new PhyNodeProp(a.getX(), a.getY());
		id = onet.nextNodeID();
		addr = a;
		net = onet;
		name = sname;
		
		physical_neighbors = new Vector();
		physical_distances = new Vector();
		physical_edges = new Vector();
		nodeInfo = new NodeInfo();
		
		timeLastMsg = net.simTime();
		//MAC = new SimpleMAC(this);
		if(sname!="legend")
			wl_MAC = new AdhocMAC(this, SimConfig.WIRELESS_DEFAULT_BANDWIDTH);

		wireless_neighbors = new Vector();
		wireless_distances = new Vector();
		wireless_sensing_neighbors = new Vector();
		
		initStats();
	}

	/**constructor
	 * used when adding via gui
	 * @param a node's physical address
	 * @param onet a reference to the overlay network
	 * @param sname node's name
	 */
	public PhysicalNode(PhyNodeProp pnode, PhysicalNetwork onet, String sname) {
		//phyNode = pnode;
		id = onet.nextNodeID();
		addr = new XYAddress(pnode.GetX(), pnode.GetY());
		net = onet;
		name = sname;

		physical_neighbors = new Vector();
		physical_distances = new Vector();
		physical_edges = new Vector();
		timeLastMsg = net.simTime();
		//MAC = new SimpleMAC(this);

		nodeInfo = new NodeInfo();
		
		initStats();
		if(pnode.wireless) {
			wl_MAC = new AdhocMAC(this, SimConfig.WIRELESS_DEFAULT_BANDWIDTH);
			statsPro.addStatsElement(wl_MAC.getStatsName(), wl_MAC, 1, 1);
		}

		wireless_neighbors = new Vector();
		wireless_distances = new Vector();
		wireless_sensing_neighbors = new Vector();

		setMobility(pnode.mobility, pnode.speed);
		setLinkType(pnode.wired, pnode.wireless);

	}
	//attributes method
	/** @return node name */
	public String getName() {
		return name;
	}
	/** set the name of this node */
	public void setName(String sname) {
		name = sname;
	}
	/** @return the physical address of the node */
	public XYAddress getAddress() {
		//return new XYAddress(phyNode.GetX(),phyNode.GetY()); 
		return addr;
	}
	/** set the address of the node */
	public void setAddress(XYAddress a) {
		//phyNode.SetXF(a.getX());
		//phyNode.SetYF(a.getY());
		addr = a;
	}

	//logical & physical links' related method
	/**
	 * test if the node is its physical neighbor
	 */
	boolean hasWiredEdge(PhysicalNode n) {
		if (physical_neighbors.contains(n))
			return true;
		else
			return false;
	}
	boolean hasPhyNeighbor(PhysicalNode n) {
		if (hasWiredEdge(n))
			return true;
		if (wireless_neighbors.contains(n))
			return true;
		return false;
	}


	/**
	 * insert a physical edge for this node
	 * @param n the new neigbhor
	 * @param weight the length of the node
	 */
	/*
	boolean insertWiredEdge(PhysicalNode n, float weight) {
		if (!wired_link)
			return false;

		if (physical_neighbors.contains(n))
			return false;
		else {
			physical_neighbors.addElement(n);
			physical_distances.addElement(new Float(weight));
			
		}
		return true;
	}*/
	/**
	 * insert a physical edge for this node
	 * 
	 */
	boolean insertWiredEdge(PhysicalNode n, WiredEdge edge) {
		if (!wired_link)
			return false;

		if (physical_neighbors.contains(n))
			return false;
		else {
			physical_neighbors.addElement(n);
			physical_distances.addElement(new Float(edge.distance));
			physical_edges.addElement(edge);
			statsPro.addStatsElement(edge.getStatsName(), edge, 1, 1);
		}
		return true;
	}
	/**
	 * @return all the physical neighbors of this node
	 */
	Enumeration getPhyNeighbor() {
		//return physical_neighbors.elements();
		//return all neighbors instead
		return getAllNeighbors().elements();
	}

	/**
	 * @return the total number of physical neighbors
	 */
	int getPhyNeighborNumber() {
		//return physical_neighbors.size(); 
		//reutrn getAllNeighbors() count instead
		return getAllNeighbors().size();
	}
	/**
	 * @return the index of this physical node in all the neighbors
	 */
	int getPhyNeighborIndex(PhysicalNode onode) {
		//return physical_neighbors.indexOf( onode );
		//reutrn getAllNeighbors() instead
		return getAllNeighbors().indexOf(onode);
	}

	/**
	 * @return the legth of a physical edge, which is represented by
	 * its index among all the neighbors
	 */
	float getPhyEdgeLen(int index) {
		return ((Float)getAllDistances().elementAt(index)).floatValue();
	}

	/**
	 * remove an physical edge
	 * @param n the node at the other end of the edge
	 * @return true if such a neighbor exists, false otherwise
	 */
	boolean removeWiredEdge(PhysicalNode n) {
		if (physical_neighbors.contains(n)) {
			int index = physical_neighbors.indexOf(n);
			physical_neighbors.remove(index);
			physical_distances.remove(index);
			WiredEdge edge = (WiredEdge)physical_edges.remove(index);
			//statsPro.removeStatsElement("Wire", index);
			statsPro.removeStatsElement(edge);
			return true;
		} else {
			return false;
		}
	}

	//routing related functions

	/**
	 * clear the routing table of this node
	 */
	public void clearRoutingTable() {
		routing_nexthop = null;
		routing_nexthop = new PhysicalNode[net.getPhyNodes().size()];
	}

	/**
	 * insert an routing table entry
	 */
	public void insertIntoRoutingTable(
		PhysicalNode destination,
		PhysicalNode next_hop,
		float distance) {
		int index = net.getPhyNodes().indexOf(destination);
		routing_nexthop[index] = next_hop;
	}

	/**
	 * @return the next hop node from this node to destination
	 */
	public PhysicalNode getPhyNextHop(PhysicalNode destination) {
		int index = net.getPhyNodes().indexOf(destination);
		//lotusDebut.println("inidex at: " + index );
		if (index < 0)
			return null;
		if (routing_nexthop == null)
			return null;
		if (routing_nexthop[index] == null)
			return null;
		return routing_nexthop[index];
	}

	
	/**
	 * part of the physicalIniterface implementation
	 * @return the physical neighbors of this node
	 */
	public Enumeration physicalLinks() {
		return getPhyNeighbor();
	}

	/** <0 if the physical node is not a neighbor
	 *  otherwise, the link weight
	 */
	//float physicalWeight( physicalInterface );

	/**
	 * @the physical address of this node
	 */
	public XYAddress physicalAddress() {
		return getAddress();
	}


	//both the timer and the messages
	/**
	 * send a message, used by the logical part of the node
	 */
	public boolean sendMsg(SimEventMsg msg) {
		if(!net.isRunning())
			return false;
		/* Overlay socket forwarding uses the same message. so it becomes troublesome when multiple sockets receive
		 * the same message object, then change that message.
		 */
		if(msg instanceof SimEventApplMsg)
			msg = (SimEventMsg)msg.clone(this);
		
		boolean sent = false;

		//lotosDebug.errorln(this.toString()+"going to send: "+msg+" to "+msg.getDst());
		msg.setSender(this);
		PhysicalNode dst = msg.getDst();
		if(dst==this) {
			//lotosDebug.print("x");
			return false;
			/*
			msg.setTime(net.simTime()+1);
			msg.setTransType(SimEventMsg.TRANS_WIRED);			
			msg.setSrc(this);
			msg.setNextHop(this);
			net.insertEvent(msg);
			sent = true;
			*/
		}
		//multicast message
		else if (dst == null) {
			//if hop limit = 0, then there is no broadcast mode
			if(net.multicast_hop_limit == 0)
				return false;
			msg.setTTL(this.net.multicast_hop_limit);
			if (wired_link) {
				//broadcast wired message
				//lotosDebug.errorln("src neighbor size="+physical_neighbors.size());
				
				for (int i = 0; i < physical_neighbors.size(); i++) {
					SimEventMsg wmsg = (SimEventMsg)msg.clone(this);
					wmsg.setTransType(SimEventMsg.TRANS_WIRED);
					wmsg.setSender(this);
					wmsg.setSrc(this);
					wmsg.setTTL(msg.getTTL());
					
					PhysicalNode wdst = (PhysicalNode)physical_neighbors.get(i);
					//float wdist = getWiredEdgeLen(i);
					float wdist = this.getDirectDistanceTo(wdst);
					
					
					//wmsg.setDst(wdst);
					wmsg.setNextHop(wdst);
					
					WiredEdge edge = (WiredEdge)physical_edges.get(i);
					edge.addMessage(wmsg, net.seq.getSimTime());
					sent = true;
					/*
					//this control structure makes no sense. they're exactly the same minus comments
					if (wmsg.eventType() == SimEvent.MESSAGE_EVENT) {
						//net.phyLayer.sendMsg(msg);
						//lotosDebug.errorln("sent: "+msg);
						wmsg.setTime(net.simTime() + wdist / PACKET_SPEED);
						net.insertEvent(wmsg);
						//MAC.sendMsg(wmsg);
						sent = true;
						//lotosDebug.errorln("sent wired broadcasting to "+wdst+":"+wmsg);
					} else {
						wmsg.setTime(net.simTime() + wdist / PACKET_SPEED);
						net.insertEvent(wmsg);
						sent = true;
					}
					*/
				}

			}

			if (wireless_link) {
				//broadcast wireless message
				msg.setTransType(SimEventMsg.TRANS_WIRELESS);

				msg.setSender(this);
				msg.setSrc(this);
				msg.setNextHop(null);
				if (msg.eventType() == SimEvent.MESSAGE_EVENT) {
					//net.phyLayer.sendMsg(msg);
					//lotosDebug.errorln("sent: "+msg);
					//adhocmac.sendmsg changes the time
					//msg.setTime(net.simTime()+ wireless_trans_radius/WirelessPhysicalLayer.PACKET_SPEED);
					//net.insertEvent( msg );
					wl_MAC.sendMsg(msg);
					sent = true;
				} else {//when does this happen?
					//msg.setTime(net.simTime()+ wireless_trans_radius/ WirelessPhysicalLayer.PACKET_SPEED);
					//net.insertEvent( msg );
					wl_MAC.sendMsg(msg);
					sent = true;
				}
			}
		}
		//unicast message
		else {
			PhysicalNode nextHop = this.getPhyNextHop(dst);

			//if(!forwarding && nextHop!=dst)return;
			boolean wired_msg = nextHop!=null && this.isWiredNeighbor(nextHop);
			
			if (nextHop==null && (!wired_msg && SimConfig.AD_HOC_ROUTING)) {
				lotosDebug.errorln("PhysicalNode.sendMsg: packet can not be routed between " + net.nodes.indexOf(this)+" and "+net.nodes.indexOf(dst));
				//return false;
			}
			float dist;

			

			if (wired_msg) {
				msg.setTransType(SimEventMsg.TRANS_WIRED);
				//dist = getWiredEdgeLen(nextHop);
			} else {
				if(wl_MAC==null)
					return false;
				msg.setTransType(SimEventMsg.TRANS_WIRELESS);
				dist = wl_MAC.SIGNAL_RANGE;
			}

			//lotosDebug.errorln(this.toString()+" is sending msg to "+nextHop);
			//lotosDebug.errorln("Message: "+msg+" to "+msg.getDst());

			msg.setSender(this);
			msg.setSrc(this);
			msg.setNextHop(nextHop);
			if (msg.eventType() == SimEvent.MESSAGE_EVENT) {
				//net.phyLayer.sendMsg(msg);

				if (wired_msg) {
					WiredEdge edge = (WiredEdge)physical_edges.get(physical_neighbors.indexOf(nextHop));
					edge.addMessage(msg, net.seq.getSimTime());
					/*
					msg.setTime(net.simTime() + dist / PACKET_SPEED);
					net.insertEvent(msg);
					*/
				} else {
					if(wl_MAC != null)
					//msg.setTime(net.simTime()+ dist / WirelessPhysicalLayer.PACKET_SPEED);
						wl_MAC.sendMsg(msg);
				}
				sent = true;
			} else {
				lotosDebug.println("??? msg is of type SimEventMsg, so this should never happen");
				//msg.setTime(net.simTime() + dist / PACKET_SPEED);
				net.insertEvent(msg);
				sent = true;
			}
		}

		return sent;

	}

	/**
	 * set a timer event
	 * @param interval the interval of this node
	 * @param index the type of the timer
	 *
	 */
	/*
	public void setTimer(long interval, Object index) {
		SimEventTimer timer =
			new SimEventTimer(net.simTime() + interval, this, index);
		//added by Guimin
		timer.setSender(this);
		timer.setNextHop(this);
		net.insertEvent(timer);
	}
	*/
	

	/**
	 * @return the simulation time
	 */
	public long simTime() {
		return net.simTime();
	}

	//message handling from/to Simulation Event Processor
	/**
	 * handle the events from simulation event
	 */
	public void handleEvent(SimEvent event) {
		try{
			if (event.eventType() == SimEvent.MESSAGE_EVENT) {
	
				
				net.sum_sent++;
	
				SimEventMsg msg = (SimEventMsg)event;
				//lotosDebug.errorln(this.toString()+" Receive: "+msg.srcaddr);
				if(!isNeighbor(msg.sender))
					lotosDebug.errorln("recieved msg from nonneighbor");
				if(msg.isWired()) {
					if(physical_neighbors.indexOf(msg.getSender())>=0) {
						WiredEdge edge = (WiredEdge) physical_edges.get(physical_neighbors.indexOf(msg.getSender()));
						edge.packetsRec++;
						edge.bytesRec += msg.msgSize;
					} 
					
				}
	
				if (msg.isWireless() && wl_lastMsgCollision) {
					lotosDebug.errorln("collision. message dropped. time=" + net.simTime() + " source="+msg.srcaddr + " dest="+msg.dstaddr);
					net.sum_collision++;
					return;
				}
				if(msg.isWireless()) {
					if(msg.msgType == SimEventMsg.APPLICATION_MESSAGE)
						msg = (SimEventMsg)msg.clone(msg.src);
					wl_MAC.bytesRec += msg.getSize();
					wl_MAC.packetsRec++;
				}				
	
				if (!(net.getPhyNodes().contains(msg.getSrc()))) {
					lotosDebug.errorln("unknown sender/receiver packet");
					return;
				}
				if (msg.getDst() == this || msg.getDst() == null) {
					
					if(msg.getTTL()>0)msg.setTTL(msg.getTTL()-1);
					//multicast forwarding
					if (forwarding && msg.getDst() == null && msg.getTTL()>0) {
						
						//lotosDebug.errorln(""+this+"forwarding multicast");
						
						//only forward multicast message to wired neighbors
						Vector phyneighbors = this.getAllNeighbors(); //wireless routing depends on table creation
						//Vector phyneighbors = this.getWiredNeighbors(); //no ad hoc wireless multicasting
						//lotosDebug.errorln("neighbor size="+this.getPhyNeighborNumber());
						
						for (int i = 0; i < phyneighbors.size(); i++) {
							
							PhysicalNode pn = (PhysicalNode)phyneighbors.elementAt(i);
							PhysicalNode nextHop = pn.getPhyNextHop(msg.getSrc()); //returns null if not in routing table
							
							//lotosDebug.errorln("test neighbor "+pn);
							/** this reads all the physical neighbors. if the routing from the
							 * neighbor to the source is this node, then this node is the parent
							 * of the neighbor, and should broadcast to that neighbor
							 */
							
							if (nextHop == this) {
								//forward the multicast data
								SimEventMsg newmsg =(SimEventMsg)msg.clone(msg.src);
								
								//lotosDebug.errorln(this.toString()+"forwarding multicast from "+msg.getSender()+" to "+pn);
								if (this.isWiredNeighbor(pn))
									newmsg.setTransType(SimEventMsg.TRANS_WIRED);
								else
									newmsg.setTransType(SimEventMsg.TRANS_WIRELESS);
	
								newmsg.setSender(this);
								newmsg.setNextHop(pn);
								newmsg.setTTL(msg.getTTL());
	
								if (this.isWiredNeighbor(pn)) { //should always be true.
									//newmsg.setTime(net.simTime()+ getDirectDistanceTo(pn)/ PACKET_SPEED);
									//net.insertEvent(newmsg);
									WiredEdge edge = (WiredEdge)physical_edges.get(physical_neighbors.indexOf(pn));
									//lotosDebug.println("forwarding multicast");
									edge.addMessage(newmsg, net.simTime());
	
								} else { 
									//adhocmac.sendmsg changes the time anyways
									//newmsg.setTime(net.simTime() + getDirectDistanceTo(pn)/ WirelessPhysicalLayer.PACKET_SPEED);
									wl_MAC.sendMsg(newmsg);
	
								}
							}
						}
					}
					//forward to the simhcnode
					if (protocol != null) {
						//lotosDebug.println("sending msg to protocol");
						protocol.handleMsg(msg);
					}
	
				} else { //unicast forwarding
					//lotosDebug.errorln("receive message from "+msg.getSrc()+" to "+msg.getDst());
					if(msg.getTTL()>0)msg.setTTL(msg.getTTL()-1);
					if (!forwarding || msg.getTTL()==0 || (msg.isWireless() && !SimConfig.AD_HOC_ROUTING))
						return;
	
					PhysicalNode nextHop = getPhyNextHop(msg.getDst());
					if (null == nextHop) {
						lotosDebug.errorln("packet can not be routed - forwarding: node:"+this.id+" src: "+msg.getSrc().id+" dst:"+msg.getDst().id);
						return;
					} else {
						if (this.isWiredNeighbor(nextHop))
							msg.setTransType(SimEventMsg.TRANS_WIRED);
						else
							msg.setTransType(SimEventMsg.TRANS_WIRELESS);
	
						//lotosDebug.errorln(this.toString()+"forwarding from "+msg.getSender()+" to "+nextHop);
						msg.setSender(this);
						msg.setNextHop(nextHop);
	
						if (this.isWiredNeighbor(nextHop)) {
							//msg.setTime(net.simTime()+ getDirectDistanceTo(nextHop) / PACKET_SPEED);
							//net.insertEvent(msg);
							WiredEdge edge = (WiredEdge)physical_edges.get(physical_neighbors.indexOf(nextHop));
							edge.addMessage(msg, net.simTime());
	
						} else {
							//msg.setTime(net.simTime()+ getDirectDistanceTo(nextHop)/ WirelessPhysicalLayer.PACKET_SPEED);
							wl_MAC.sendMsg(msg);
	
						}
					}
				}
			}
		}catch(Exception e){
			lotosDebug.errorln("Exception when handling event: " + event);
			e.printStackTrace();
		}
	}

	/**
	 * resets all physical edges
	 * 
	 */
	void resetWiredEdge() {
		Enumeration edges = physical_edges.elements();
		while(edges.hasMoreElements()) {
			((WiredEdge)edges.nextElement()).clearMsgs();
		}
	}
	/** set the logical node running on this node */
	void setProtocol(I_ProtocolInterface p) {
		if(protocol!=null)
			statsPro.removeStatsElement(protocol);
			
		protocol = p;
		if(p!=null)
			statsPro.addStatsElement(protocol.getStatsName(), protocol, 1, 1);
	}

	/** puts a hypercast node into the statistics */
	/*
	public void registerHCNode(I_Stats hcnode) {
		if(nodestats)
			statsPro.removeStatsElement("Node", 0);
		statsPro.addStatsElement("Node", hcnode, 1, 1);
		nodestats = true;
	}
	*/
	/** return the logical node running on this node*/
	public I_ProtocolInterface getProtocol() {
		return protocol;
	}

	public String toString() {
		if (getProtocol() == null)
			return "@" + addr;
		else
			return getProtocol().toString();
	}
	/* (non-Javadoc)
	 * @see physicalInterface#getTimer(int)
	 */
	public long getTimer(Object index) {
		return net.getEvent(new SimEventTimer(0, this, index));
	}
	/* (non-Javadoc)
	 * @see physicalInterface#clearTimer(int)
	 */
	public void clearTimer(Object index) {
		net.clearEvent(new SimEventTimer(0, this, index));
	}
	public boolean equals(Object o){
		return this == o;
	}
	
	private void initStats() {
		statsPro = new StatsProcessor(this/* I_Stats object */, true/* readable flag */, true/* writable flag */);
		statsPro.addStatsElement("NodeID", new stat_nodeID(), 1, 1);
		statsPro.addStatsElement("Location", new stat_location(), 1, 1);
		statsPro.addStatsElement("TotalPacketsSent", new stat_packetsSent(), 1, 1);
    	statsPro.addStatsElement("TotalBytesSent", new stat_bytesSent(), 1, 1);
    	statsPro.addStatsElement("TotalPacketsRecieved", new stat_packetsRec(), 1, 1);
    	statsPro.addStatsElement("TotalBytesRecieved", new stat_bytesRec(), 1, 1);
    	
	}
	class stat_location extends SimpleStats {
		
		protected String getStats() {
			return addr.toString();
		}
		
		protected String setStats(String newValue) throws HyperCastStatsException {
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}
		
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException {
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:String"/* type */, 
					null/* restriction base */, null/* pattern value */);
			
		}
	}
	class stat_nodeID extends SimpleStats {
			
		protected String getStats() {
			return Integer.toString(id);
		}
		
		protected String setStats(String newValue) throws HyperCastStatsException {
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}
		
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException {
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
					null/* restriction base */, null/* pattern value */);
			
		}
	}
	
	class stat_bytesSent extends SimpleStats{
		protected String getStats()
		{
			int i;
			long bytes=0;
			for(i=0;i<physical_edges.size();i++) {
				bytes+=((WiredEdge)physical_edges.get(i)).bytesSent;
			}
			return Long.toString(bytes);
		}

		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}
	class stat_packetsSent extends SimpleStats{
		protected String getStats()
		{
			int i;
			long packets=0;
			for(i=0;i<physical_edges.size();i++) {
				packets+=((WiredEdge)physical_edges.get(i)).packetsSent;
			}
			return Long.toString(packets);
		}

		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}
	class stat_bytesRec extends SimpleStats{
		protected String getStats()
		{
			int i;
			long bytes=0;
			for(i=0;i<physical_edges.size();i++) {
				bytes+=((WiredEdge)physical_edges.get(i)).bytesRec;
			}
			return Long.toString(bytes);
		}

		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}
	class stats_packetsRec extends SimpleStats{
		protected String getStats()
		{
			int i;
			long packets=0;
			for(i=0;i<physical_edges.size();i++) {
				packets+=((WiredEdge)physical_edges.get(i)).packetsRec;
			}
			return Long.toString(packets);
		}

		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}
	/** see Monitor and Control API */
	class stat_packetsRec extends SimpleStats{
		protected String getStats()
		{
			int i;
			long packets=0;
			for(i=0;i<physical_edges.size();i++) {
				packets+=((WiredEdge)physical_edges.get(i)).packetsRec;
			}
			return Long.toString(packets);
		}

		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}
	public String getStatsName() {
		return statsName;
	}
	public void setStatsName(String name) {
		statsName = name;
	}
	public Element[] getStats(Document doc, XPath xpath) throws HyperCastStatsException {
		return statsPro.getStatsResult(doc, xpath);
	}
	public Element[] setStats(Document doc, XPath xpath, Element newValue) throws HyperCastStatsException {
		return statsPro.setStatsResult(doc, xpath, newValue);
	}
	public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException {
		return statsPro.getReadSchemaResult(doc, xpath);
	}
	public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException {
		return statsPro.getWriteSchemaResult(doc, xpath);
	}

}
