package lotos;

/**
 * SimControl is the control interface for the rest of lotos. It can
 * a. modify the simulation attributes
 * b. advance the simulation
 * c. access the statistics XML tree of lotos
 * 
 * the statistics of lotos can control every facet of the simulator:
 * a. modify the physical network
 * b. create and destroy hypercast nodes
 * c. create and destroy lotos gateways
 * d. display the current status of the simulator and simulation
 * e. record and display simulation data
 * 
 * @author Zhi An
 *
 */

import org.apache.xpath.XPath;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import lotos.*;
import lotos.event.*;
import hypercast.*;
import hypercast.util.XmlUtil;

import java.util.*;

public class SimControl implements I_Stats {
	public simHCNetwork network;
	public boolean running;
	
	public SimControl() {
		running = false;
		initStats();
	}
	
	public boolean initNetwork() {
		if(network!=null)
			return false;
		network = new simHCNetwork();
		if(SimConfig.log !=null && !SimConfig.log.logging())
			SimConfig.log.startLog(new Long(System.currentTimeMillis()).toString());
		
		statsPro.addStatsElement(network.statsName, network, 1, 1);
		return true;
	}
	public boolean restartNetwork() {
		if(network == null)
			return false;
		if(SimConfig.log !=null) {
			SimConfig.log.endLog();
			SimConfig.log.startLog(new Long(System.currentTimeMillis()).toString());
		}
		network.restart();
		return true;
	}
	public boolean resetNetwork() {
		if(network==null)
			return false;
		if(SimConfig.log !=null)
			SimConfig.log.endLog();
		statsPro.removeStatsElement(network);
		network.stop();
		network = null;
		initNetwork();

		SimEventProcessor.reset();
		return true;
	}

	public boolean start() {
		if(SimConfig.log !=null && !SimConfig.log.logging())
			SimConfig.log.startLog(new Long(System.currentTimeMillis()).toString());
		return network.start();
	}
	
    /**
     * run the simulation for the specific amount of time
     * @param time the duration to run the simulation in microseconds
     */
	public final void advance(long adv_time)
	{		
		if(network == null)
			return;
		long sim_time=SimEventProcessor.getSimTime();
		long big_end_time=sim_time+adv_time;
		long end_time;
		while(sim_time<big_end_time){
			SimEvent se;
			end_time = sim_time + SimConfig.SIM_STEP;//<adv_time?SIM_STEP:adv_time;
			//lotosDebug.println(""+end_time);
			if(network.hasMobileNode) {
				network.moveMobileNodes(); //moves mobile wireless nodes
				network.updateWirelessAdjacency(); //updates wireless neighbors
			}
			
			network.buildPhysicalRoute(); //creates routing tables
			
			//network.MACStep(); does nothing
			network.checkPhyLinks();
			
			while( (se=SimEventProcessor.getNextEvent(end_time)) != null )
			//while( (se=getNextEvent(end_time)) != null )
			{
				//lotosDebug.println("event");
				//setMSSimTime(se.getMSTime());
				sim_time=se.getTime();
				//lotosDebug.println(""+se.getRecipient());
				//lotosDebug.println(""+se.eventType());
				if(se.target==null && se instanceof SimEventMsg) //this should happen only in broadcast wireless
				{
					SimEventMsg msg = (SimEventMsg) se;
					//se.getSender().receiveEvent(se);
					for(int i =0;i <msg.getSender().getWirelessNeighbors().size();i++)
					{
						PhysicalNode temp = (PhysicalNode)msg.getSender().getWirelessNeighbors().elementAt(i);
						
						//lotosDebug.errorln("sent to one recipient");
						
						if(temp!=msg.getSender())
							temp.handleEvent(se);
					}
				}
				else if(se.target != null)
				{
					se.target.handleEvent(se);
				} else {
					lotosDebug.errorln("Unknown event encountered" + se.toString());
				}
				//sim_time=se.time;
				//lotosDebug.errorln(""+sim_time+":"+se);

				network.checkPhyLinks();
			}
			//setSimTime(end_time);
			sim_time = end_time;
		}
		//lotosDebug.errorln("advance finished");
		
	}
	
	 /**
     * advance the simulation to the next pending event
     */
	public final void stepToNext()
	{
		if(SimEventProcessor.queueIsEmpty())return;
		
		SimEvent se;
		long sim_time = SimEventProcessor.getSimTime();
		long end_time = sim_time + SimConfig.SIM_STEP;

		while(end_time > sim_time)
		{	
			if(network.hasMobileNode) {
				network.moveMobileNodes(); //moves mobile wireless nodes
				network.updateWirelessAdjacency(); //updates wireless neighbors
			}
			
			network.buildPhysicalRoute(); //creates routing tables
			
			//network.MACStep(); does nothing
			network.checkPhyLinks();
			
			if(SimEventProcessor.queueIsEmpty())return;
			
			while( (se=SimEventProcessor.getNextEvent(end_time)) != null )
			{
				//sim_time=se.getTime();
				sim_time=se.getTime();
				
				if(se.target==null && se instanceof SimEventMsg)
				{
					SimEventMsg msg = (SimEventMsg) se;
					//se.getSender().receiveEvent(se);
					for(int i =0;i <msg.getSender().getWirelessNeighbors().size();i++)
					{
						PhysicalNode temp = (PhysicalNode)msg.getSender().getWirelessNeighbors().elementAt(i);
						if(temp!=msg.getSender())
							temp.handleEvent(se);
					}
				}
				else if(se.target != null) {
					se.target.handleEvent(se);
				}
				
				if((se.eventType() == SimEvent.MESSAGE_EVENT)){
					//lotosDebug.errorln(""+sim_time+":"+se);
					return;
				}
				
			}
			
			//setSimTime(end_time);
			sim_time = end_time;
			end_time = sim_time + SimConfig.SIM_STEP;
			
		}
	}
	
	String statsName="Lotos";
	StatsProcessor statsPro;
	
	private void initStats() {
		statsPro = new StatsProcessor(this/* I_Stats object */, true/* readable flag */, true/* writable flag */);
		statsPro.addStatsElement("AdvanceSteps", new stat_advStep(), 1, 1);
		statsPro.addStatsElement("AdvanceTime", new stat_advTime(), 1, 1);
	}
	class stat_advStep extends SimpleStats
	{
		
		protected String getStats()
		{
			return Integer.toString(0);
		}

		protected String setStats(String newValue)
		{
			int numSteps = Integer.valueOf(newValue).intValue();
			if(network == null || numSteps<0)
				return newValue;
			if(!network.isRunning())
				network.start();
			advance(SimConfig.SIM_STEP*numSteps);
						
			return new String("0");
		}

		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
		
		public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException {
			return XmlUtil.createSchemaElement(doc, statisticsName, "xsd:Integer", null, null);
		}
	}
	class stat_advTime extends SimpleStats
	{
		
		protected String getStats()
		{
			return Integer.toString(0);
		}

		protected String setStats(String newValue)
		{
			long time = (long)(Float.valueOf(newValue).floatValue() * SimEventProcessor.TIME_PER_SEC);
			if(network == null || time<0)
				return newValue;
			if(!network.isRunning())
				network.start();
			advance(time);
						
			return new String("0");
		}

		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
		
		public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException {
			return XmlUtil.createSchemaElement(doc, statisticsName, "xsd:Integer", null, null);
		}
	}
	public String getStatsName() {
		return statsName;
	}
	public void setStatsName(String name) {
		statsName = name;
	}
	public Element[] getStats(Document doc, XPath xpath) throws HyperCastStatsException {
		return statsPro.getStatsResult(doc, xpath);
	}
	public Element[] setStats(Document doc, XPath xpath, Element newValue) throws HyperCastStatsException {
		return statsPro.setStatsResult(doc, xpath, newValue);
	}
	public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException {
		return statsPro.getReadSchemaResult(doc, xpath);
	}
	public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException {
		return statsPro.getWriteSchemaResult(doc, xpath);
	}
}
