/***************************************************************************/
/*       Lotos, an extensible overlay network simulator                    */
/*       MNG Group                                                         */
/*       University of Virginia                                            */
/*       July, 2002                                                        */
/***************************************************************************/
package lotos;

import lotos.*; 
import lotos.event.SimEvent;

import java.util.*;

/**
 * The discrete-time simulator
 */
/*
 * revised by Guangyu Dong, Jan 6, 2004
 * to get precise microsecond level simulation time
 */
/*
 * changed to static. this clock is integral in all of lotos. having it global makes for more readable and less
 * circuitous code. specifically, all the intermediary steps (which were really just redirections) to add/remove
 * events are removed. 
 * 
 * zhi an, june 7, 2006
 */
public class SimEventProcessor
{
    public final static float TIME_PER_SEC=1000000f;
    
    /** current simulation time */
    /** changed to "long" by Guangyu to get precise microsecond level time */
    /** 1000000 = 1sec */
    private static long sim_time;
    //float sim_time;
    
    /* set sim time with microseconds */
    public static void setSimTime(long t){
    	sim_time=t;	
    }
    
    /* @return current simulation time in microseconds */
    public static long getSimTime(){
    	return sim_time;	
    }
        
    /** the queue containing the simulation event */
    static Vector the_queue;
	
	/** Constructor */
    public SimEventProcessor()
    {
        setSimTime(0);
        the_queue = new Vector();
    }

    /** @return those pending events */
    public synchronized static Enumeration getPendingEvents() {
        return ((Vector)the_queue.clone()).elements();
    }


    /** insert an event into the simulator */
    public synchronized static void insertEvent( SimEvent se)
    {
    	
        if (se.getTime() < getSimTime())
        {
        	/*
            lotosDebug.error("ERROR: Event inserted for time " + se.getTime() );
            //lotosDebug.error(se.getTime());
            lotosDebug.error(" when clock was at " + sim_time );
            //lotosDebug.error(sim_time);
            lotosDebug.errorln(": Event Ignored.");
            return;
            */
        	se.setTime(getSimTime()+1);
        }
        
        //added by Guangyu, Feb 10th, 2004
        //to remove duplicated timer event
		clearEvent(se);
        int i;
        for( i=0; i<the_queue.size(); i++){
            SimEvent event = (SimEvent) the_queue.elementAt(i);
            if( event.getTime() > se.getTime() ) break;
        }
        if( i == the_queue.size() ) the_queue.addElement(se);
        else the_queue.add( i, se );
    }

    /** @return true when the event queue is empty */
    public synchronized static boolean queueIsEmpty()
    {
        if (the_queue.size() == 0 ) return true; else return false;
    }

    /** @return the simulation time of the next event in microseconds, -1 if the queue is empty */
    public synchronized static long getNextTime() {
    	if(the_queue.size()==0)
    		return -1;
    	SimEvent se = (SimEvent) the_queue.elementAt(0);
    	return se.getTime();
    }


    /** @return the next pending event and advance the sim time */
    public synchronized static SimEvent getNextEvent()
    {
        if( the_queue.size() ==  0 ) return null;
        SimEvent se = (SimEvent) the_queue.elementAt(0);
        the_queue.remove(0);
        setSimTime(se.getTime()); //advance the time;
        return se;
    }

    /** @return the next event within the endTime
     *  If the return is null, advance the time to end time
     *  If the return is not-null, advance the time using getNextEvent
     */

    public synchronized static SimEvent getNextEvent(long endTime ){

        if( the_queue.size() == 0 ) {
            setSimTime(endTime);
            return null;
        }
        SimEvent se = (SimEvent) the_queue.elementAt(0);
        if( se.getTime() > endTime ) { setSimTime(endTime); return null; }
        return getNextEvent();
    }

    /** reset the discrete-time simulator */
    public synchronized static void reset(){
        the_queue.clear();
        sim_time = 0;
    }
    /**
     * Find a specified type of event
     * @param e		the event
     * @return		the delay of the event, or -1
     */
    public synchronized static long getEvent(SimEvent e){
		//lotosDebug.errorln("get a event"+e);
    	for(int i=0;i<the_queue.size();i++){
    		SimEvent se=(SimEvent)the_queue.elementAt(i);
    		if(se.eventType()==e.eventType() && se.equals(e)){
    			//lotosDebug.errorln("find a event"+se);
    			return se.getTime()-getSimTime();
    		}
    	}
    	return -1;
    }
    /**
     * Cleare a specified type of event
     * @param e		the event
     */
    public synchronized static void clearEvent(SimEvent e){
    	//if(!(e instanceof SimEventTimer))return;
    	
		for(int i=0;i<the_queue.size();i++){
			SimEvent se=(SimEvent)the_queue.elementAt(i);
			if(se.eventType()==e.eventType() && se.equals(e))
				the_queue.removeElementAt(i);
		}
	
    }
    

}
