/***************************************************************************/
/*       Lotos, an extensible overlay network simulator                    */
/*       MNG Group                                                         */
/*       University of Virginia                                            */
/*       July, 2002                                                        */
/***************************************************************************/
package lotos;

import lotos.*;
import lotos.event.SimEvent;
import lotos.event.SimEventMsg;
import hypercast.*;
import hypercast.util.XmlUtil;

import java.util.LinkedList;

import org.apache.xpath.XPath;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
/**
 * This class implements an edge in the physical network. It is mainly
 * used by the topology adaptors to store the topology description information.
 * 
 * The physical edges handles the simulation event messages sent between nodes.
 * It simulates the physical network connection between two nodes.
 * 
 * The simulation event timer messages required by the hypercast protocols is
 * handled by SimEventProcessor.
 * 
 * @author Zhi An
 */

public class WiredEdge implements I_Stats{

    /** the source node */
	public PhysicalNode srcNode;
    /** the destination node */
	public PhysicalNode dstNode;
    /** the length of the edge */
    public float distance;
    /** the bandwidth of the edge in bytes / second */
    public int bandwidth;
    /** the list of messages queued by the physical edge to be sent */
    public LinkedList msgs;
    /** the size of the send buffer in bytes. if the buffer is exceeded, then the oldest
      * messages are deleted until the newest message fits. if the message's size > bufferSize,
      * the message is ignored.
      */
    int bufferSize;
    /** the buffer space that is in use */
    int usedBuffer; 
    /** The next time a message will be finished being transmitted, in microseconds
     *  Latency isn't included in this.
     *  finishSending == -1 if the edge is idle, i.e. no message is being transmitted.
     * */
    public long finishSending;
    
    /** statistics */
    StatsProcessor statsPro;
    String statsName = "Wire";
    long packetsSent;
    long bytesSent;
    long packetsRec;
    long bytesRec;
    
    public WiredEdge( PhysicalNode src, PhysicalNode dst, float dist, int bw, int bufsize){
    	srcNode = src;
    	dstNode = dst;
    	distance = dist;
    	bufferSize = bufsize;
    	bandwidth = bw;
    	msgs = new LinkedList();
    	usedBuffer = 0;
    	finishSending = -1;
    	
    	packetsSent = 0;
    	bytesSent = 0;
    	packetsRec = 0;
    	bytesRec = 0;

    	initStats();
    }
    /** creates a new physical edge with source and destination reversed. useful for 
     * symmetric connections
     */
    public WiredEdge reverse() {
    	return new WiredEdge(dstNode, srcNode, distance, bandwidth, bufferSize);
    }
    /** add message, returns the time the message will be recieved by the recipient 
     *	if the buffer is exceeded, then the first message in the queue is deleted
     *  until enough of the buffer is empty to add the new message. if the size of 
     *  the new message + the size of the message currently being sent > buffer size,
     *  then the new message is ignored. 
     *  
     *  @param msg the message to be inserted
     *  @param currentTime the current simulation time in microseconds
     *  @return true if the message is successfully added to the edge, false otherwise
     */ 
    public synchronized boolean addMessage(SimEventMsg msg, long currentTime) {
    	/* the message is only buffered if the edge is not currently sending a message.
    	 * if the message is bigger than the buffer, then discard the message. if all
    	 * the messages are larger than the buffer size, then the edge acts as if it doesn't
    	 * have a buffer, and will only add messages if it is idle.
    	 */ 
    	if(msgs.size()>0 && msg.msgSize + ((SimEventMsg)msgs.getFirst()).msgSize > bufferSize)
    		return false;
    	/*
    	 * clear the buffer until the message fits. the above check assures that the message will
    	 * fit into the buffer, i.e. the buffer can be sufficiently cleared
    	 * the oldest message in the buffer is removed first.
    	 */
	    while(msgs.size()>1 && usedBuffer + msg.msgSize > bufferSize) {
			//lotosDebug.println("buffer exceeded");
			SimEventMsg tmp = (SimEventMsg)msgs.remove(1); //the first one is currently being sent
			usedBuffer -= tmp.msgSize;
		}
	    
    	long sendTime = getSendTime(msg);
    	
    	if(msgs.size() == 0) { //the edge is currently idle
    		msgs.add(msg);
    		finishSending = currentTime + sendTime;

    	} else {
    		msgs.add(msg);
    		usedBuffer += msg.msgSize;

    	}
    	return true;
    }
    
    /** 
     * getMessage returns the message that is being sent on the edge
     * @param currentTime the current simulation time
     * @return the event that is currently being sent, null if the edge is idle
     */
    public synchronized SimEvent getMessage(){
    	if(msgs.size() == 0)
    		return null;
    	SimEventMsg msg = (SimEventMsg) msgs.removeFirst();
    	msg.setTime(finishSending + getLatency());
    	/*
    	 * if there are messages in the send buffer, then promote the oldest to being sent.
    	 */
    	if(msgs.size() > 0) {
	    	SimEventMsg tmp = (SimEventMsg) msgs.getFirst();
	    	long sendTime = getSendTime(tmp);
	    	finishSending += sendTime;
	    	usedBuffer -= tmp.msgSize;
	    	
    	} else { //the buffer is empty, set the edge to idle
    		finishSending = -1;
    	}
    	packetsSent++;
    	bytesSent += msg.msgSize;
    	
    	return msg;
    }
    /**
     * deletes all the messages on the edge
     *
     */
    public synchronized void clearMsgs() {
    	msgs = new LinkedList();
    	finishSending = -1;
    	usedBuffer = 0;
    }
    /**
     * resets the edge to its original condition
     *
     */
    public void reset() {
    	clearMsgs();
    	packetsSent = 0;
    	bytesSent = 0;
    	packetsRec = 0;
    	bytesRec = 0;
    }
    
    /**
     * this calculates the time it takes for the bandwidth of the edge to transmit the message
     *  @return the time in microseconds required for the message to be transmitted
     */
    public long getSendTime(SimEventMsg msg) {
    	return (long)(msg.msgSize * SimEventProcessor.TIME_PER_SEC / bandwidth);
    }
    /**
     * 
     * @return the latency of the edge
     */
    public long getLatency() {
    	return (long)(distance / SimConfig.WIRE_PACKET_SPEED * SimEventProcessor.TIME_PER_SEC);
    }
    
    /**
     * changes the bandwidth of the edge
     * @param bw the new bandwidth, bw must be > 0
     * @return true if successful, false if failed
     */
    public boolean setBandwidth(int bw) {
    	if(bw<=0)
    		return false;
    	bandwidth = bw;
    	return true;
    }
    /**
     * changes the length of the wired edge
     * @param newDistance the new length
     */
    public void setDistance(float newDistance) {
    	distance = newDistance>0?newDistance:0-newDistance;
    }
    
    /**
     * changes the buffer size of the edge. 
     * @param buf the new buffer size, must be > 0
     * @return true if successful, false if failed
     */
    public boolean setBuffer(int buf) {
    	if(buf<0)
    		return false;
    	clearMsgs();
    	return true;
    }
    
    /**
     * statistics section
     * r values: packetsSent,packetsRec,bytesSent,bytesRec, usedBuffer
     * r/w values: bandwidth, bufferSize
     *
     */
    private void initStats() {
    	statsPro = new StatsProcessor(this/* I_Stats object */, true/* readable flag */, true/* writable flag */);
    	statsPro.addStatsElement("DestinationNode", new stat_destNode(), 1, 1);
    	statsPro.addStatsElement("PacketsSent", new stat_packetsSent(), 1, 1);
    	statsPro.addStatsElement("BytesSent", new stat_bytesSent(), 1, 1);
    	statsPro.addStatsElement("PacketsRecieved", new stat_packetsRec(), 1, 1);
    	statsPro.addStatsElement("BytesRecieved", new stat_bytesRec(), 1, 1);
    	statsPro.addStatsElement("Length", new stat_distance(), 1, 1);
    	statsPro.addStatsElement("Bandwidth", new stat_bandwidth(), 1, 1);
    	statsPro.addStatsElement("BufferSize", new stat_bufferSize(), 1, 1);
    	statsPro.addStatsElement("BufferUsed", new stat_usedBuffer(), 1, 1);
    	statsPro.addStatsElement("PacketsBuffered", new stat_packetsBuffed(), 1, 1);
    	
    }
    class stat_distance extends SimpleStats
	{
		
		protected String getStats()
		{
			return Float.toString(distance);
		}

		protected String setStats(String newValue)
		{
			float newdist = Float.valueOf(newValue).floatValue();
			if(newdist>=0)
				distance = newdist;
			return Float.toString(distance);	
		}

		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Float"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
		public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException {
			return XmlUtil.createSchemaElement(doc, statisticsName, "xsd:Float", null, null);
		}
	}
    class stat_destNode extends SimpleStats
	{
		
		protected String getStats()
		{
			return Integer.toString(dstNode.net.nodes.indexOf(dstNode));
		}

		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Int"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}
    class stat_packetsSent extends SimpleStats
	{
		
		protected String getStats()
		{
			return Long.toString(packetsSent);
		}

		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}
    class stat_bytesSent extends SimpleStats
	{
		
		protected String getStats()
		{
			return Long.toString(bytesSent);
		}

		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}
    class stat_packetsRec extends SimpleStats
	{
		
		protected String getStats()
		{
			return Long.toString(packetsRec);
		}

		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}
    class stat_bytesRec extends SimpleStats
	{
		
		protected String getStats()
		{
			return Long.toString(bytesRec);
		}

		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}
    class stat_usedBuffer extends SimpleStats
	{
		
		protected String getStats()
		{
			return Long.toString(usedBuffer);
		}

		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}
    class stat_bandwidth extends SimpleStats
	{
		
		protected String getStats()
		{
			return Integer.toString(bandwidth);
		}

		protected String setStats(String newValue)
		{
			int newBand = Integer.valueOf(newValue).intValue();
			setBandwidth(newBand);
			
			return Integer.toString(bandwidth);	
		}

		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
		public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException {
			return XmlUtil.createSchemaElement(doc, statisticsName, "xsd:Integer", null, null);
		}
	}
    class stat_bufferSize extends SimpleStats
	{
		
		protected String getStats()
		{
			return Integer.toString(bufferSize);
		}

		protected String setStats(String newValue)
		{
			int newbuf = Integer.valueOf(newValue).intValue();
			setBuffer(newbuf);
			
			return Integer.toString(bufferSize);
		}

		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
		
		public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException {
			return XmlUtil.createSchemaElement(doc, statisticsName, "xsd:Integer", null, null);
		}
	}
    class stat_packetsBuffed extends SimpleStats
	{
		
		protected String getStats()
		{
			return Integer.toString(msgs.size()>0?msgs.size()-1:0);
		}

		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Int"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}
    /** see Monitor and Control API */
    public String getStatsName() {
		return statsName;
	}
	public void setStatsName(String name) {
		statsName = name;
	}
	public Element[] getStats(Document doc, XPath xpath) throws HyperCastStatsException {
		return statsPro.getStatsResult(doc, xpath);
	}
	public Element[] setStats(Document doc, XPath xpath, Element newValue) throws HyperCastStatsException {
		return statsPro.setStatsResult(doc, xpath, newValue);
	}
	public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException {
		return statsPro.getReadSchemaResult(doc, xpath);
	}
	public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException {
		return statsPro.getWriteSchemaResult(doc, xpath);
	}
}