/***************************************************************************/
/*       Lotos, an extensible overlay network simulator                    */
/*       MNG Group                                                         */
/*       University of Virginia                                            */
/*       July, 2002                                                        */
/***************************************************************************/
package lotos.event;

import lotos.*;
import hypercast.I_Message;
import hypercast.OL_Message;
import hypercast.SecInfoExchange_Message;
/**
 * This class implements message event by extending SimEvent
 * 
 * it carries generic transmission information and size in bytes. 
 */

/*
 * if the message has a unicast target, use setNextHop to set the handler for the event.
 */
public abstract class SimEventMsg extends SimEvent
{
	/* Variables pertaining to the encapsulated hypercast message */
	/** message type */
	public final static int PROTOCOL_MESSAGE=0;	
	public final static int APPLICATION_MESSAGE=1;
	
	/** the actual message from hypercast */
	public I_Message hc_message;
	
    /** type of the message */
    public int msgType;
    
    /** size of the message. defaults to 0, messages and data can change this value */
    public int msgSize = 0;
	
    /* variables pertaining to physical layer routing */
    /** source of the message, note that src can be different from sender in the
     *  SimEvent because a message can be forwarded for multiple hops before
     *  reaching the destination. While src and dst of the message are the same
     *  during the whole process, sender and recipient are changed at each forwarding
     *  node
     */
	public PhysicalNode src;
    /** destination of the message */
    public PhysicalNode dst;

    public SimPhysicalAddress srcaddr;
    public SimPhysicalAddress dstaddr;
    
    /** current hop information */
    public PhysicalNode sender;
    public PhysicalNode nextHop;
    
    /** added by Guangyu */
    /** TTL */
    int TTL=-1;
    
    /** type of transmission */
    int transType;
    public final static int TRANS_WIRED=0;
    public final static int TRANS_WIRELESS=1;
    public final static int TRANS_BOTH=2;
  
    public SimEventMsg(PhysicalNode osrc, PhysicalNode odst, int mtype, I_Message hc_msg )
    {
        super( 0, null, SimEvent.MESSAGE_EVENT );
        src = osrc;
        dst = odst;
        msgType = mtype;
        synchronized(hc_msg) {
        	msgSize = hc_msg.toByteArray().length;
        }
        hc_message = hc_msg;
    }
    
    public void setTransType(int tt){
    	transType=tt;	
    }
    public boolean isWired(){
    	return transType==TRANS_WIRED || transType==TRANS_BOTH;
    }
    public boolean isWireless(){
    	return transType==TRANS_WIRELESS || transType==TRANS_BOTH;
    }
	
	public int getTTL(){
		return TTL;
	}
	public void setTTL(int ttl){
		TTL=ttl;
	}



    public PhysicalNode getSrc()
    {
        return src;
    }

    public void setSrc( PhysicalNode o)
    {
        src = o;
    }
    //public void setSrc( PhysicalNode osrc ) { src = osrc; }

    public PhysicalNode getDst()
    {
        return dst;
    }

    public void setDst( PhysicalNode o)
    {
        dst = o;
    }

    //public void setDst( PhysicalNode odst ) { dst = odst; }
    public int getMsgType(){ return msgType; }

	public int getSize(){ return msgSize; }
    
    /** current hop */
    public PhysicalNode getNextHop()
    {
        return nextHop;
    }

    public PhysicalNode getSender()
    {
        return sender;
    }

    public void setNextHop(PhysicalNode onode ){
        nextHop = onode;
        setTarget(nextHop);
    }

    public void setSender(PhysicalNode onode ){
        sender = onode;
    }
    
	public String getMsgLabel() {
		return hc_message.toString();
	}

	public Object clone(I_PhysicalInterface phy) {
		LotosAdapter target;
		I_Message clone;
		if(msgType==APPLICATION_MESSAGE) {
			clone = new OL_Message((OL_Message)hc_message);
		} else {
			target = phy.getProtocol().getNodeAdapter();
			int[] valid = {0};
			clone = target.callback.restoreMessage(hc_message.toByteArray(), valid, msgSize);
			if(clone instanceof SecInfoExchange_Message) {
				((SecInfoExchange_Message)(clone)).setProtoMsg(((SecInfoExchange_Message)hc_message).getProtoMsg());
			}
		}
		
		SimEventMsg msg;
		if(msgType==APPLICATION_MESSAGE)
			msg = new SimEventApplMsg(src, dst, clone);
		else
			msg = new SimEventProtocolMsg(src, dst, clone);
		msg.srcaddr = srcaddr;
		msg.dstaddr = dstaddr;
		msg.setSender(sender);
		msg.setNextHop(nextHop);
		msg.transType = transType;
		
		return msg;
	}
    /**
	 * @return	the I_Message object contained in this class
	 */
	public I_Message getHCMessage(){
		return hc_message;
	}
}
